import numpy as np


class Indexer(object):
    '''
    Base indexer class for accessing subsets of time series ensemble data.
    
    Provides utility methods for indexing 2D coordinate data (x, y) which is
    common in wavefront sensor measurements and other spatial data.
    
    This class serves as a foundation for specialized indexers like
    :class:`ModeIndexer`, :class:`DefaultIndexer`, and :class:`RowColIndexer`.
    '''

    def __init__(self):
        self.dicCoord = {'x': 0, 'y': 1}

    def _xy_index(self, coord):
        if not hasattr(coord, '__iter__'):
            coord = [coord]
        indexCoord = [self.dicCoord[x] for x in coord]
        return indexCoord

    def interleaved_xy(self, *args, **kwargs):
        '''
        Create slice for interleaved x/y data layout.
        
        This indexer is designed for wavefront sensor slopes stored in
        interleaved format: [x0, y0, x1, y1, x2, y2, ...]
        
        Parameters
        ----------
        *args : str or int, optional
            If 'x': select x coordinates (indices 0, 2, 4, ...)
            If 'y': select y coordinates (indices 1, 3, 5, ...)
            If int: select single element at that index
            If not provided: select all elements
        coord, coords, axis : {'x', 'y'}, optional
            Alternative keyword-based coordinate selection.
        
        Returns
        -------
        slice
            Python slice object for indexing the requested coordinates.
        
        Examples
        --------
        >>> indexer = Indexer()
        >>> slopes = np.array([sx0, sy0, sx1, sy1, sx2, sy2])  # interleaved
        >>> 
        >>> # Select all x slopes
        >>> x_slopes = slopes[indexer.interleaved_xy('x')]  # [sx0, sx1, sx2]
        >>> 
        >>> # Select all y slopes  
        >>> y_slopes = slopes[indexer.interleaved_xy(coord='y')]  # [sy0, sy1, sy2]
        '''
        acceptedKeywords = ['coord', 'coords', 'axis']
        for k, v in kwargs.items():
            if k not in acceptedKeywords:
                raise Exception('possible keywords are ' +
                                str(acceptedKeywords))
            if v == 'x':
                return slice(0, None, 2)
            elif v == 'y':
                return slice(1, None, 2)
            else:
                raise Exception("%s not in accepted"
                                " values ('x', 'y')" % (v))

        # Since we are here, no keywords were specified
        if len(args) == 0:
            return slice(0, None, 1)
        elif len(args) == 1:
            if args[0] == 'x':
                return slice(0, None, 2)
            elif args[0] == 'y':
                return slice(1, None, 2)
            elif isinstance(args[0], int):
                return slice(args[0], args[0]+1)
        else:
            raise Exception("%s not in accepted"
                            " values ('x', 'y')" % str(v))

    def sequential_xy(self, maxindex, *args, **kwargs):
        '''
        Create slice for sequential x/y data layout.
        
        This indexer is for data stored with all x coordinates followed by
        all y coordinates: [x0, x1, x2, ..., y0, y1, y2, ...]
        
        Parameters
        ----------
        maxindex : int
            Total number of elements (length of the array to index).
        *args : str or int, optional
            If 'x': select x coordinates (first half)
            If 'y': select y coordinates (second half)
            If int: select single element at that index
            If not provided: select all elements
        coord, coords, axis : {'x', 'y'}, optional
            Alternative keyword-based coordinate selection.
        
        Returns
        -------
        slice
            Python slice object for indexing the requested coordinates.
        
        Examples
        --------
        >>> indexer = Indexer()
        >>> slopes = np.array([sx0, sx1, sx2, sy0, sy1, sy2])  # sequential
        >>> 
        >>> # Select all x slopes
        >>> x_slopes = slopes[indexer.sequential_xy(6, 'x')]  # [sx0, sx1, sx2]
        >>> 
        >>> # Select all y slopes
        >>> y_slopes = slopes[indexer.sequential_xy(6, coord='y')]  # [sy0, sy1, sy2]
        '''
        acceptedKeywords = ['coord', 'coords', 'axis']
        for k, v in kwargs.items():
            if k not in acceptedKeywords:
                raise Exception('possible keywords are ' +
                                str(acceptedKeywords))
            if v == 'x':
                return slice(0, maxindex//2)
            elif v == 'y':
                return slice(maxindex//2, None)
            else:
                raise Exception("%s not in accepted"
                                " values ('x', 'y')" % (v))

        # Since we are here, no keywords were specified
        if len(args) == 0:
            return slice(0, None, 1)
        elif len(args) == 1:
            if args[0] == 'x':
                return slice(0, maxindex//2)
            elif args[0] == 'y':
                return slice(maxindex//2, None)
            elif isinstance(args[0], int):
                return slice(args[0], args[0]+1)
        else:
            raise Exception("%s not in accepted"
                            " values ('x', 'y')" % str(v))
        
    def xy(self, *args, **kwargs):
        '''
        default: coord
        Accepted keywords:
        coord= x and/or y
        '''
        acceptedKeywords = ['coord', 'coords', 'axis']
        coord = ['x', 'y']
        if len(args) > 0:
            coord = args[1]
        for k, v in kwargs.items():
            if k not in acceptedKeywords:
                raise Exception('possible keywords are ' +
                                str(acceptedKeywords))
            if k == 'coord' or k == 'coords' or k == 'axis':
                coord = v
        return self._xy_index(coord)

    # TODO add boundary checks with maxrange on args
    @staticmethod
    def myrange(*args, maxrange):
        
        if len(args)==0:
            return range(maxrange)
        if len(args)==1:
            return range(args[0])
        if len(args)==2:
            return range(args[0], args[1])
        if len(args)==3:
            return range(args[0], args[1], args[2])
        raise ValueError
            
class ModeIndexer(Indexer):
    '''
    Indexer for modal decomposition data (Zernike, KL modes, etc.).
    
    Provides convenient indexing for time series containing modal coefficients,
    allowing selection by mode number, mode ranges, or lists of specific modes.
    
    Parameters
    ----------
    max_mode : int, optional
        Maximum mode number available in the data. Used as default upper bound
        for range selections.
    
    Examples
    --------
    >>> # For Zernike coefficients time series with 100 modes
    >>> indexer = ModeIndexer(max_mode=100)
    >>> 
    >>> # Select all modes
    >>> all_modes = indexer.modes()  # [0, 1, 2, ..., 99]
    >>> 
    >>> # Select mode range (tip, tilt, focus, astig)
    >>> low_order = indexer.modes(from_mode=2, to_mode=6)  # [2, 3, 4, 5]
    >>> 
    >>> # Select specific modes
    >>> selected = indexer.modes(modes=[2, 5, 10])  # [2, 5, 10]
    '''

    def __init__(self, max_mode=None):
        super().__init__()
        self._max_mode = max_mode        

    def modes(self, *args, **kwargs):
        '''
        Generate mode indices for selection.
        
        Parameters
        ----------
        *args : int or array-like, optional
            Single mode number or list of mode numbers to select.
        mode, modes : int or array-like, optional
            Alternative keyword-based mode selection.
        from_mode : int, optional
            Starting mode number for range selection. Default is 0.
        to_mode : int, optional
            Ending mode number (exclusive) for range selection.
            If not specified, uses `max_mode` from initialization.
        
        Returns
        -------
        int, list, or ndarray
            Mode index/indices for array indexing.
        
        Raises
        ------
        Exception
            If `to_mode` is not specified and `max_mode` was not set
            during initialization.
        
        Examples
        --------
        >>> indexer = ModeIndexer(max_mode=50)
        >>> 
        >>> # All modes
        >>> indexer.modes()  # array([0, 1, 2, ..., 49])
        >>> 
        >>> # Single mode
        >>> indexer.modes(mode=5)  # 5
        >>> 
        >>> # Mode range (e.g., low-order Zernikes)
        >>> indexer.modes(from_mode=2, to_mode=11)  # array([2, 3, ..., 10])
        >>> 
        >>> # Specific modes
        >>> indexer.modes(modes=[2, 3, 4, 10, 15])  # [2, 3, 4, 10, 15]
        '''
        modes = None
        from_mode = None
        to_mode = None
        if (len(args) > 0) & (len(args) < 2):
            modes = args[0]
        for k, v in kwargs.items():
            if k == 'modes' or k == 'mode':
                modes = v
            if k == 'from_mode':
                from_mode = v
            if k == 'to_mode':
                to_mode = v
        if modes is not None:
            return modes
        if from_mode is None:
            from_mode = 0
        if to_mode is None:
            to_mode = self._max_mode
        if to_mode is None:
            raise Exception("to_mode was not specified and max_mode was not set "
                            "when initializing the ModeIndexer, cannot continue") 
        return np.arange(from_mode, to_mode)


class DefaultIndexer(Indexer):
    '''
    Generic element indexer for uniform 1D ensemble data.
    
    Provides flexible indexing for time series where ensemble elements
    are not organized in a special structure (modes, rows/cols, etc.)
    but simply form a flat list.
    
    Examples
    --------
    >>> indexer = DefaultIndexer()
    >>> 
    >>> # Select all elements
    >>> indexer.elements()  # slice(None, None, None)
    >>> 
    >>> # Select single element
    >>> indexer.elements(22)  # 22
    >>> 
    >>> # Select list of elements
    >>> indexer.elements([2, 5, 10])  # [2, 5, 10]
    >>> 
    >>> # Select range
    >>> indexer.elements(from_element=10, to_element=20)  # slice(10, 20, None)
    '''

    def __init__(self):
        super().__init__()

    def elements(self, *args, **kwargs):
        '''
        Generate element indices for selection.
        
        Parameters
        ----------
        *args : int, list, or array-like, optional
            Single element index or list of indices to select.
        element, elements : int, list, or array-like, optional
            Alternative keyword-based element selection.
        from_element : int, optional
            Starting element index for range selection.
        to_element : int, optional
            Ending element index for range selection.
        
        Returns
        -------
        int, list, or slice
            Index/indices for array selection.
        
        Examples
        --------
        >>> indexer = DefaultIndexer()
        >>> 
        >>> # All elements (no filtering)
        >>> indexer.elements()  # slice(None, None)
        >>> 
        >>> # Single element
        >>> indexer.elements(5)  # 5
        >>> 
        >>> # List of elements
        >>> indexer.elements([1, 3, 5, 7])  # [1, 3, 5, 7]
        >>> 
        >>> # Range from index 10 onwards
        >>> indexer.elements(from_element=10)  # slice(10, None)
        '''
        elements = None
        from_element = None
        to_element = None
        if (len(args) > 0) & (len(args) < 2):
            elements = args[0]
        for k, v in kwargs.items():
            if k == 'elements' or k == 'element':
                elements = v
            if k == 'from_element':
                from_element = v
            if k == 'to_element':
                to_element = v
        if elements is not None:
            return elements
        return slice(from_element, to_element)


class RowColIndexer(Indexer):
    '''
    Indexer for 2D array data organized in rows and columns.
    
    Useful for time series data with natural 2D spatial structure,
    such as deformable mirror actuators arranged in a grid or
    detector pixels.
    
    Examples
    --------
    >>> indexer = RowColIndexer()
    >>> 
    >>> # Select specific rows and columns
    >>> idx = indexer.rowcol(rows=5, cols=slice(0, 10))
    >>> data_subset = data[:, idx[0], idx[1]]  # time x rows x cols
    >>> 
    >>> # Using from/to syntax
    >>> idx = indexer.rowcol(row_from=10, row_to=20, col_from=5, col_to=15)
    '''

    def __init__(self):
        super().__init__()

    def rowcol(self, *args, **kwargs):
        '''
        Create row and column indices for 2D data selection.
        
        Parameters
        ----------
        *args : int, slice, or array-like
            Positional arguments for row and column selection:
            - If 1 arg: used for rows, all columns selected
            - If 2 args: (rows, cols)
        rows : int, slice, or array-like, optional
            Row index/indices to select.
        cols : int, slice, or array-like, optional
            Column index/indices to select.
        row_from, row_to, row_step : int, optional
            Convenience parameters for building row slice.
        col_from, col_to, col_step : int, optional
            Convenience parameters for building column slice.
        
        Returns
        -------
        tuple of (row_index, col_index)
            Indices suitable for 2D array indexing.
        
        Raises
        ------
        IndexError
            If unsupported keywords or too many positional arguments provided.
        
        Examples
        --------
        >>> indexer = RowColIndexer()
        >>> 
        >>> # Select row 5, columns 0-10
        >>> rows, cols = indexer.rowcol(rows=5, col_from=0, col_to=10)
        >>> 
        >>> # Using positional arguments
        >>> rows, cols = indexer.rowcol(slice(10, 20), slice(5, 15))
        >>> 
        >>> # Select all rows, specific column range with step
        >>> rows, cols = indexer.rowcol(col_from=0, col_to=100, col_step=2)
        '''
        rows = slice(None, None, None)
        cols = slice(None, None, None)
        if len(args) == 1:
            rows = args[0]
        elif len(args) == 2:
            rows, cols = args
        elif len(args) > 2:
            raise IndexError('Unsupported rowcol index')
        
        accepted = 'rows cols row_from col_from row_to col_to row_step col_step'.split()
        for kw in kwargs:
            if kw not in accepted:
                raise IndexError(f'Unsupported index keyword {kw}')

            if kw == 'rows':
                rows = kwargs['rows']
            if kw == 'cols':
                cols = kwargs['cols']
            if kw == 'row_from':
                rows = slice(kwargs['row_from'], rows.stop, rows.step)
            if kw == 'col_from':
                cols = slice(kwargs['col_from'], cols.stop, cols.step)
            if kw == 'row_to':
                rows = slice(rows.start, kwargs['row_to'], rows.step)
            if kw == 'col_to':
                cols = slice(cols.start, kwargs['col_to'], cols.step)
            if kw == 'row_step':
                rows = slice(rows.start, rows.stop, kwargs['row_step'])
            if kw == 'col_step':
                cols = slice(cols.start, cols.stop, kwargs['col_step'])

        return (rows, cols)



import numpy as np
from numbers import Number


class ModalCoefficients():
    FIRST_MODE = 0

    def __init__(self, coefficients, counter=0, first_mode=0):
        self._coefficients = coefficients
        self._counter = counter
        self.FIRST_MODE = first_mode

    def modeIndexes(self):
        return np.arange(self.FIRST_MODE, self.FIRST_MODE + self.numberOfModes())

    def numberOfModes(self):
        return len(self._coefficients)

    def getM(self, modeIndexes):
        return self.toNumpyArray()[np.array(modeIndexes) - self.FIRST_MODE]

    def toDictionary(self):
        keys = self.modeIndexes()
        values = self._coefficients
        return dict(list(zip(keys, values)))

    def toNumpyArray(self):
        return self._coefficients

    @staticmethod
    def fromNumpyArray(coefficientsAsNumpyArray, counter=0, **kwargs):
        return ModalCoefficients(np.array(coefficientsAsNumpyArray), counter, **kwargs)

    def counter(self):
        return self._counter

    def setCounter(self, counter):
        self._counter = counter

    def __eq__(self, o):
        if self._counter != o._counter:
            return False
        if not np.array_equal(self._coefficients, o._coefficients):
            return False
        return True

    def __ne__(self, o):
        return not self.__eq__(o)

    def __str__(self):
        return str(self._coefficients)

    def __add__(self, other):
        if isinstance(other, ModalCoefficients):
            if len(self._coefficients) < len(other._coefficients):
                c = other._coefficients.copy()
                c[:len(self._coefficients)] += self._coefficients
            else:
                c = self._coefficients.copy()
                c[:len(other._coefficients)] += other._coefficients
            return ModalCoefficients(c)
        if isinstance(other, Number):
            return ModalCoefficients(self._coefficients + other)
        return NotImplemented

    def __radd__(self, other):
      return self.__add__(other)

    def __iadd__(self, other):
        if isinstance(other, ModalCoefficients):
            if len(self._coefficients) < len(other._coefficients):
                c = other._coefficients.copy()
                c[:len(self._coefficients)] += self._coefficients
            else:
                c = self._coefficients.copy()
                c[:len(other._coefficients)] += other._coefficients
            self._coefficients = c
            return self
        elif isinstance(other, Number):
            self._coefficients += other
            return self
        return NotImplemented

    def __neg__(self):
        return ModalCoefficients(-self._coefficients)

    def __pos__(self):
        return ModalCoefficients(self._coefficients)

    def __abs__(self):
        return pow(sum(coo**2 for coo in self._coefficients), 0.5)

    def __sub__(self, other):
        return self + (-other)

    def __rsub__(self, other):
        return other + (-self)

    def __isub__(self, other):
        self += -other
        return self

    def __mul__(self, other):
        if isinstance(other, Number):
            return ModalCoefficients(self._coefficients * other)
        return NotImplemented

    def __rmul__(self, other):
        return self * other

    def __imul__(self, other):
        if isinstance(other, Number):
            self._coefficients *= other
            return self
        return NotImplemented

    def __truediv__(self, other):
        if isinstance(other, Number):
            return ModalCoefficients(self._coefficients / other)
        return NotImplemented

    def __rtruediv__(self, other):
        if isinstance(other, Number):
            return ModalCoefficients(other / self._coefficients)
        return NotImplemented

    def __itruediv__(self, other):
        if isinstance(other, Number):
            self._coefficients /= other
            return self
        return NotImplemented



"""Data elaboration framework for adaptive optics laboratory data.

This module provides a framework for analyzing time-series data from adaptive
optics systems. It implements a pattern-based architecture with three main
components working together:

1. **FileWalker**: Manages directory structure and file paths
2. **DataLoader**: Implements lazy loading of data from files
3. **TimeSeries**: Wraps time-series data with metadata and analysis methods
4. **Analyzer**: Coordinates data loading and provides analysis interface

The framework is designed for datasets organized by "tags" (unique identifiers,
typically timestamps) where each tag contains:

- Snapshot of system state (slow-varying parameters)
- Time series of fast-varying data (camera frames, DM commands, slopes)
- Calibration data (reconstructor matrices, reference frames)

Key Features
------------

- **Lazy loading**: Data loaded only when accessed
- **Disk caching**: Computed results cached to disk for performance
- **Unit handling**: Automatic unit conversion using astropy.units
- **Flexible indexing**: Select data subsets (e.g., x/y slopes, quadrants)
- **Batch analysis**: Analyze multiple tags together via AnalyzerSet

Basic Usage
-----------

1. Derive from AbstractFileNameWalker to define your file structure
2. Create TimeSeries subclasses for your data types
3. Derive from BaseAnalyzer to create your main analyzer
4. Use Analyzer.get(tag) to load and analyze data

See Also
--------
arte.dataelab.base_analyzer : Main analyzer base class
arte.dataelab.base_timeseries : Time series base class
arte.dataelab.base_file_walker : File walker base class
"""


import numpy as np
from abc import abstractmethod
import matplotlib.pyplot as plt


class Coronograph(object):
    """ Abstract class to simulate coronographic PSFs """
    
    @abstractmethod
    def _get_pupil_mask(self, field):
        """ Override this method with the function
        returning the coronograph pupil-plane mask """

    @abstractmethod
    def _get_focal_plane_mask(self, field):
        """ Override this method with the function
        returning the coronograph focal-plane mask """

    def _get_apodizer(self):
        """ Override this method if the coronograph
        has an apodizer mask """
        return 1.0

    def get_coronographic_psf(self, input_field, oversampling:int, lambdaInM:float=None):
        """
        Function to obtain the post-coronographic PSF.

        Parameters
        ----------
        input_field: numpy.ndarray(complex) [N,N]
            Array of complex values representing the pupil-plane
            electric field upstream the coronograph.
            - Real part: electric field amplitude
            - Imaginary part: electric field phase in RADIANS

        oversampling: int
            Integer for the number of pixels to consider for each lambda/D.

        lambdaInM: float (None)
            Wavelength in meters at which the PSF is evaluated.
            This must be the same as the wavelength at which the 
            input_field phase was computed.
            Default is None and the coronograph uses it reference
            wavelength: i.e. we assume that the coronograph was
            optimized for the wavelength at which the input field 
            is given.

        Returns
        -------
        psf: numpy.ndarray(float) [N*oversampling,N*oversampling]
            The array representing the (padded) PSF.
            Each pixel in the array corresponds to oversampling lambda/D
        """
        self.oversampling = oversampling
        self.lambdaInM = lambdaInM
        pad_width = int(max(input_field.shape)*(self.oversampling-1))//2
        self._apodizer = self._get_apodizer()
        apodized_field = input_field * self._apodizer
        padded_field = np.pad(apodized_field,pad_width=pad_width,mode='constant',constant_values=0.0)
        self._focal_field = np.fft.fftshift(np.fft.fft2(padded_field))
        self._focal_mask = self._get_focal_plane_mask(self._focal_field)
        prop_focal_field = self._focal_field * self._focal_mask
        pupil_field = np.fft.ifft2(np.fft.ifftshift(prop_focal_field))
        self._pupil_mask = self._get_pupil_mask(padded_field)
        coro_field = pupil_field * self._pupil_mask
        self._focal_coro_field = np.fft.fftshift(np.fft.fft2(coro_field))
        psf = abs(self._focal_coro_field)**2
        return psf
    
    def show_coronograph_prop(self, maxLogPsf=None):
        """ Simple function to show the EF propagation through the coronograph """
        phase = np.angle(self._focal_mask)
        fcmap = 'RdBu'
        phase += 2*np.pi * (phase < 0.0)
        if np.max(phase) <= 1e-12:
            fcmap = 'grey'
            phase = self._focal_mask.copy()
        plt.figure(figsize=(22,4))
        plt.subplot(1,4,1)
        plt.imshow(phase,cmap=fcmap,origin='lower')
        plt.title('Focal plane mask')
        plt.colorbar()
        plt.subplot(1,4,2)
        self.showZoomedPSF(np.abs(self._focal_field)**2,
                           1/self.oversampling,title='PSF at focal mask',
                           maxLogVal=maxLogPsf)
        plt.subplot(1,4,3)
        plt.imshow(np.abs(self._pupil_mask),cmap='grey',origin='lower')
        plt.title('Pupil stop')
        plt.subplot(1,4,4)
        self.showZoomedPSF(np.abs(self._focal_coro_field)**2,
                           1/self.oversampling,title='Coronographic PSF',
                           maxLogVal=maxLogPsf)

    @staticmethod
    def showZoomedPSF(image, pixelSize, maxLogVal = None, title='',
                   xlabel=r'$\lambda/D$', ylabel=r'$\lambda/D$', zlabel=''):
        imageHalfSizeInPoints= image.shape[0]/2
        roi= [int(imageHalfSizeInPoints*0.8), int(imageHalfSizeInPoints*1.2)]
        imageZoomedLog= np.log(image[roi[0]: roi[1], roi[0]:roi[1]])
        if maxLogVal is None:
            maxLogVal = np.max(imageZoomedLog)
        imageZoomedLog -= maxLogVal
        sz=imageZoomedLog.shape
        plt.imshow(imageZoomedLog, 
                extent=[-sz[0]/2*pixelSize, sz[0]/2*pixelSize,
                    -sz[1]/2*pixelSize, sz[1]/2*pixelSize],
                    cmap='twilight',vmin=-24)
        plt.title(title)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        cbar= plt.colorbar()
        cbar.ax.set_ylabel(zlabel)


from math import e
import warnings
import numpy as np
from astropy import units as u
from synphot.spectrum import SpectralElement
from synphot.models import Empirical1D
from synphot.utils import generate_wavelengths, merge_wavelengths
from astropy.io import fits


def set_element_id_from_method(func):
    """
    Decorator that sets the 'id' attribute of a returned TransmissiveElement
    to the name of the method.
    """
    def wrapper(cls, *args, **kwargs):
        element = func(cls, *args, **kwargs)
        if isinstance(element, TransmissiveElement):
            element.set_id(func.__name__)
        return element
    return wrapper


class Bandpass():

    _MIN_LAMBDA_NM = 200
    _MAX_LAMBDA_NM = 10000
    _DELTA_LAMBDA_NM = 10

    @classmethod
    def standard_waveset(cls):
        return generate_wavelengths(
            minwave=cls._MIN_LAMBDA_NM,
            maxwave=cls._MAX_LAMBDA_NM,
            delta=cls._DELTA_LAMBDA_NM,
            log=False, wave_unit=u.nm)[0]

    @classmethod
    def one(cls):
        return cls.flat(1)

    @classmethod
    def zero(cls):
        return cls.flat(0)

    @classmethod
    def flat(cls, amplitude):
        wv = cls.standard_waveset()
        return SpectralElement(
            Empirical1D, points=wv, lookup_table=amplitude * np.ones(wv.shape))

    @classmethod
    def from_array(cls, waveset, values):
        return SpectralElement(Empirical1D, points=waveset, lookup_table=values)

    @classmethod
    def ramp(cls, low_wl, low_ampl, high_wl, high_ampl):
        '''
        T=low_ampl for lambda < low_wl
        T=high_ampl for lambda > high_wl
        T=linear_interpolation(low_ampl, high_ampl) for low_wl < lambda < high_wl
        '''
        l1 = low_wl.to(u.nm).value
        l2 = high_wl.to(u.nm).value
        return SpectralElement(
            Empirical1D,
            points=np.array([cls._MIN_LAMBDA_NM, l1, l2, cls._MAX_LAMBDA_NM]
                            ) * u.nm,
            lookup_table=np.array([low_ampl, low_ampl, high_ampl, high_ampl]))

    @classmethod
    def peak(cls, peak_wl, delta_wl, high_ampl, low_ampl):
        '''
        T = high_ampl for lambda = peak_wl
        T = low_ampl for lambda < (peak_wl - delta_wl) and 
            lambda > (peak_wl + delta_wl)
        T = linear_interpolation(low_ampl, high_ampl) for 
            (peak_wl - delta_wl) < lambda < peak_wl and
            peak_wl < lambda < (peak_wl + delta_wl) 
        '''
        l = peak_wl.to(u.nm).value
        dl = delta_wl.to(u.nm).value
        return SpectralElement(
            Empirical1D,
            points=np.array([
                cls._MIN_LAMBDA_NM, l - dl, l, l + dl, cls._MAX_LAMBDA_NM]) * u.nm,
            lookup_table=np.array([
                low_ampl, low_ampl, high_ampl, low_ampl, low_ampl]))

    @classmethod
    def top_hat(cls, peak_wl, delta_wl, high_ampl, low_ampl):
        '''
        T = high_ampl for lambda = peak_wl +/- delta_wl
        T = low_ampl elsewhere
        '''
        l = peak_wl.to(u.nm).value
        dl = delta_wl.to(u.nm).value
        return SpectralElement(
            Empirical1D,
            points=np.array([
                cls._MIN_LAMBDA_NM, l - dl, l - dl + 1e-8, l + dl,
                l + dl + 1e-8, cls._MAX_LAMBDA_NM]) * u.nm,
            lookup_table=np.array([
                low_ampl, low_ampl, high_ampl, high_ampl, low_ampl, low_ampl]))

    @classmethod
    def top_hat_ramped(cls, low_wl_start, high_wl_start,
                       high_wl_end, low_wl_end,
                       low_ampl, high_ampl):
        l1 = low_wl_start.to(u.nm).value
        l2 = high_wl_start.to(u.nm).value
        l3 = high_wl_end.to(u.nm).value
        l4 = low_wl_end.to(u.nm).value
        return SpectralElement(
            Empirical1D,
            points=np.array([cls._MIN_LAMBDA_NM, l1, l2, l3, l4, cls._MAX_LAMBDA_NM]
                            ) * u.nm,
            lookup_table=np.array([low_ampl, low_ampl, high_ampl, high_ampl, low_ampl, low_ampl]))

    @classmethod
    def step(cls, wl, low_ampl, high_ampl):
        '''
        T=low_ampl for lambda < wl
        T=high_ampl for lambda > wl
        '''
        return cls.ramp(wl, low_ampl, wl + 1e-8 * u.Angstrom, high_ampl)


class TransmissiveElement():
    '''
    An optical element with transmittance, reflectance and absorptance

    Only 2 of the 3 parameters must be specified, as the sum of the 3 quantities
    must be equal to 1 by energy conservation at every wavelength.

    Absorptance and absorptivity are synonimous. The emissivity is equal to the
    absorptance under the assumptions of Kirchhoff's law of thermal radiation.
    (Absorpance is instead a quantity related to the optical depth).

    When a scalar value is specified, a constant spectrum of the specified
    amplitude is assumed

    Parameters
    ----------
        transmittance: `synphot.SpectralElement` or scalar in [0,1]
            spectral transmittance of the element (in wavelength)

        reflectance: `synphot.SpectralElement` or scalar in [0,1]
            spectral reflectance of the element (in wavelength)

        absorptance: `synphot.SpectralElement` or scalar in [0,1]
            spectral absorptance of the element (in wavelength)
    '''

    def __init__(self, transmittance=None, reflectance=None, absorptance=None, id=""):
        self._initialize(transmittance, reflectance, absorptance)
        self._id = id

    def _initialize(self, transmittance=None, reflectance=None, absorptance=None):
        t, r, a = self._check_params(transmittance, reflectance, absorptance)
        self._t = t
        self._r = r
        self._a = a

    def _check_params(self, t, r, a):
        if t is None:
            if r is None or a is None:
                raise TypeError("Only one among t,r,a can be None")
            t = self._computeThird(r, a)
            return t, r, a
        if r is None:
            if t is None or a is None:
                raise TypeError("Only one among t,r,a can be None")
            r = self._computeThird(t, a)
            return t, r, a
        if a is None:
            if t is None or r is None:
                raise TypeError("Only one among t,r,a can be None")
            a = self._computeThird(t, r)
            return t, r, a
        raise TypeError("At least one among t,r,a must be None")

    def _computeThird(self, a, b):
        wv = a.waveset
        ones = Bandpass.one()
        lookup_values = ones(wv) - a(wv) - b(wv)
        
        # Check in range [0,1]
        min_val = np.min(lookup_values)
        max_val = np.max(lookup_values)
        
        NUMERICAL_ERROR_THRESHOLD = 1e-10
        
        if min_val < -NUMERICAL_ERROR_THRESHOLD:
            warnings.warn(
                f"t+r+a=1 cannot be fulfilled: minimum value is {min_val:.2e} "
                f"(threshold: {-NUMERICAL_ERROR_THRESHOLD:.2e})")
        
        if max_val > (1.0 + NUMERICAL_ERROR_THRESHOLD):
            warnings.warn(
                f"t+r+a=1 cannot be fulfilled: maximum value is {max_val:.2e} "
                f"(threshold: {1.0 + NUMERICAL_ERROR_THRESHOLD:.2e})")
        
        # If we reach here, the errors are numerical: clip the values
        lookup_values = np.clip(lookup_values, 0, 1)
        
        c = SpectralElement(
            Empirical1D, points=wv,
            lookup_table=lookup_values)
        return c

    def plot(self, transmittance=True, reflectance=True,
             absorptance=True, wv_unit=None, **kwargs):
        import matplotlib.pyplot as plt
        if wv_unit is None:
            wv = self.waveset
        else:
            wv = self.waveset.to(wv_unit)
        if transmittance:
            plt.plot(wv, self.transmittance(wv), label='t', **kwargs)
        if reflectance:
            plt.plot(wv, self.reflectance(wv), label='r', **kwargs)
        if absorptance:
            plt.plot(wv, self.absorptance(wv), label='a', **kwargs)
        plt.legend()
        plt.grid(True)

    @property
    def waveset(self):
        return self.transmittance.waveset

    @property
    def waverange(self):
        return self.transmittance.waverange

    @property
    def transmittance(self):
        return self._t

    @property
    def absorptance(self):
        return self._a

    @property
    def reflectance(self):
        return self._r

    @property
    def emissivity(self):
        return self.absorptance

    @property
    def id(self):
        return self._id

    def set_id(self, id):
        self._id = id

    def _compute_average_in_band(self, spectral_el, wv_band, atol, ext_waveset):
        """
        Compute average of spectral element in a wavelength band using linear interpolation.
        
        This method interpolates the spectral element at the exact band limits
        rather than searching for nearest wavelengths, eliminating the need for atol.
        
        Parameters
        ----------
        spectral_el : SpectralElement
            The spectral element to average
        wv_band : tuple of Quantity
            (wv_min, wv_max) wavelength band limits
        atol : tuple
            Ignored in this implementation (kept for backward compatibility)
        ext_waveset : array-like or None
            Optional external waveset for integration
            
        Returns
        -------
        wv_min, wv_max, average : tuple
            The band limits and the average value in the band
            
        Risks
        -----
        - Extrapolation: If band limits are outside the spectral element's waveset,
          synphot will extrapolate (usually as constant). Check that wv_band is
          within the valid range of the spectral element.
        - Interpolation accuracy: Linear interpolation may not accurately represent
          sharp features in the spectrum. For spectral elements with high-frequency
          variations, ensure adequate sampling.
        """
        if ext_waveset is None:
            wv = self.waveset
        else:
            wv = ext_waveset
        
        wv_min = wv_band[0]
        wv_max = wv_band[1]
        
        # Ensure both are in the same unit
        wv_min = wv_min.to(wv.unit)
        wv_max = wv_max.to(wv.unit)
        
        # Check if band limits are within waveset bounds
        if wv_min < np.min(wv) or wv_max > np.max(wv):
            warnings.warn(
                f"Band limits [{wv_min}, {wv_max}] extend beyond waveset range "
                f"[{np.min(wv)}, {np.max(wv)}]. Extrapolation will be used. "
                f"Element id: {self.id}")
        
        # Generate dense waveset within the band for accurate integration
        # Use the original sampling if possible
        delta_wv = np.median(np.diff(wv.value))
        n_points = max(int((wv_max.value - wv_min.value) / delta_wv), 100)
        wv_band_dense = np.linspace(wv_min.value, wv_max.value, n_points) * wv.unit
        
        # Remove any potential duplicates
        wv_band_dense = np.unique(wv_band_dense)
        
        # Evaluate spectral element at dense points
        values = spectral_el(wv_band_dense)
        
        # Compute mean using trapezoidal integration
        average = np.trapezoid(values.value, wv_band_dense.to(wv.unit).value) / \
                (wv_max - wv_min).to(wv.unit).value
        
        return wv_min, wv_max, average

    def transmittance_in_band(self, wv_band, atol=(1, 1), ext_waveset=None):
        '''
        Parameters
        ----------
        wv_band: astropy.units.quantity.Quantity
            Bounds of the wavelength range where to compute the average transmittance of the TransmissiveElement.
        '''
        wv_min, wv_max, t_mean = self._compute_average_in_band(
            self.transmittance, wv_band, atol, ext_waveset)
        return t_mean, wv_min, wv_max

    def reflectance_in_band(self, wv_band, atol=(1, 1), ext_waveset=None):
        '''
        Parameters
        ----------
        wv_band: astropy.units.quantity.Quantity
            Bounds of the wavelength range where to compute the average reflectance of the TransmissiveElement.
        '''
        wv_min, wv_max, r_mean = self._compute_average_in_band(
            self.reflectance, wv_band, atol, ext_waveset)
        return r_mean, wv_min, wv_max

    def emissivity_in_band(self, wv_band, atol=(1, 1)):
        '''
        Parameters
        ----------
        wv_band: astropy.units.quantity.Quantity
            Bounds of the wavelength range where to compute the average emissivity of the TransmissiveElement.
        '''
        wv_min, wv_max, e_mean = self._compute_average_in_band(
            self.emissivity, wv_band, atol)
        print(f'Average emissivity in band {(wv_min, wv_max)}: {e_mean}')
        return e_mean

    def _insert_spectral_element(self, spectral_element, new_values, wavelengths):
        if new_values is None:
            return None
        ws = spectral_element.waveset
        ii = np.searchsorted(ws, wavelengths)
        new_waveset = np.insert(ws, ii, wavelengths)

        r_arr = spectral_element(ws)
        new_r = np.insert(r_arr, ii, new_values)
        return SpectralElement(Empirical1D,
                               points=new_waveset,
                               lookup_table=new_r)

    def add_spectral_points(self, wavelengths, transmittance=None,
                            reflectance=None, absorptance=None):
        new_t = self._insert_spectral_element(self.transmittance, transmittance,
                                              wavelengths)
        new_r = self._insert_spectral_element(self.reflectance, reflectance,
                                              wavelengths)
        new_a = self._insert_spectral_element(self.absorptance, absorptance,
                                              wavelengths)

        self._initialize(transmittance=new_t, reflectance=new_r,
                         absorptance=new_a)

    def to_dat(self, filepath, data_type):
        if data_type == 'transmittance':
            data = self.transmittance(self.waveset).value
            title = np.array([['Wavelength [um]', 'Transmittance']])
        elif data_type == 'reflectance':
            data = self.reflectance(self.waveset).value
            title = np.array([['Wavelength [um]', 'Reflectance']])
        elif data_type == 'absorptance':
            data = self.absorptance(self.waveset).value
            title = np.array([['Wavelength [um]', 'Absorptance']])
        to_write = np.stack((
            self.waveset.to(u.um).value,
            data), axis=1)
        with open(filepath, 'a') as datafile:
            np.savetxt(datafile,
                       np.vstack((title, to_write)),
                       fmt=['%-30s', '%-30s'])

    def to_fits(self, filename, **kwargs):
        wv = self.waveset
        hdu1 = fits.PrimaryHDU(wv.to(u.nm).value)
        hdu2 = fits.ImageHDU(self.transmittance(wv).value)
        hdu3 = fits.ImageHDU(self.absorptance(wv).value)
        hdul = fits.HDUList([hdu1, hdu2, hdu3])
        hdul.writeto(filename, **kwargs)

    @staticmethod
    def from_fits(filename):
        wv = fits.getdata(filename, 0) * u.nm
        t = SpectralElement(Empirical1D, points=wv,
                            lookup_table=fits.getdata(filename, 1))
        a = SpectralElement(Empirical1D, points=wv,
                            lookup_table=fits.getdata(filename, 2))
        return TransmissiveElement(transmittance=t, absorptance=a)

    @staticmethod
    def ideal():
        """
        Creates an ideal transmissive element with unit transmittance across all wavelengths.
        
        Returns
        -------
        TransmissiveElement
            A transmissive element with transmittance = 1.0 everywhere,
            reflectance = 0.0 everywhere, and absorptance = 0.0 everywhere.
        """
        return TransmissiveElement(transmittance=Bandpass.one(), 
                                   reflectance=Bandpass.zero())

    @staticmethod
    def flat(transmittance=1.0, reflectance=0.0):
        """
        Creates a transmissive element with constant values across all wavelengths.
        
        Parameters
        ----------
        transmittance : float, optional
            Constant transmittance value in [0,1]. Default is 1.0.
        reflectance : float, optional
            Constant reflectance value in [0,1]. Default is 0.0.
        
        Returns
        -------
        TransmissiveElement
            A transmissive element with constant transmittance and reflectance,
            and absorptance computed from t + r + a = 1.
            
        Raises
        ------
        ValueError
            If transmittance + reflectance > 1.
        """
        if transmittance + reflectance > 1.0:
            raise ValueError(
                f"transmittance ({transmittance}) + reflectance ({reflectance}) > 1.0")
        
        return TransmissiveElement(
            transmittance=Bandpass.flat(transmittance),
            reflectance=Bandpass.flat(reflectance))


class Direction(object):
    TRANSMISSION = 'transmission'
    REFLECTION = 'reflection'


class TransmissiveSystem():

    def __init__(self, name="Transmissive System"):
        self._name = name
        self._elements = []
        self._waveset = np.array([1000]) * u.nm

    def add(self, transmissive_element_or_system, direction=None, name=''):
        if transmissive_element_or_system is None:
            warnings.warn(
                "Trying to add a None element (named '%s') to TransmissiveSystem. Ignoring" % name)
            return
        if isinstance(transmissive_element_or_system, TransmissiveSystem):
            for el in transmissive_element_or_system.elements:
                self.add(el['element'], direction=el['direction'],
                         name=el['name'])
        else:
            if direction is None:
                raise ValueError(
                    "direction must be specified when adding a TransmissiveElement")
            self._elements.append(
                {'element': transmissive_element_or_system,
                 'name': name,
                 'direction': direction})
            self._waveset = merge_wavelengths(
                self._waveset.to(u.nm).value,
                transmissive_element_or_system.waveset.to(u.nm).value) * u.nm

    def remove(self, element_index):
        del self._elements[element_index]

    @property
    def name(self):
        return self._name

    @property
    def waveset(self):
        return self._waveset

    @property
    def transmittance(self):
        return self._compute_transmittance()

    def transmittance_from_to(self, from_element=0, to_element=None):
        return self._compute_transmittance(from_element, to_element)

    def subsystem_from_to(self, from_element=0, to_element=None):
        subsystem = TransmissiveSystem(name=self._name)
        if to_element is None:
            end_element = len(self._elements)
        else:
            end_element = to_element + 1
        for el in self._elements[from_element:end_element]:
            subsystem.add(el['element'], direction=el['direction'], name=el['name'])
        return subsystem
        

    def _compute_transmittance(self, from_element=0, to_element=None):
        t = 1
        if to_element is None:
            end_element = to_element
        else:
            end_element = to_element + 1
        for el in self._elements[from_element:end_element]:
            direction = el['direction']
            if direction == Direction.TRANSMISSION:
                t *= el['element'].transmittance
            elif direction == Direction.REFLECTION:
                t *= el['element'].reflectance
            else:
                raise ValueError(
                    'Unknown propagation direction %s' % direction)
        return t

    @property
    def emissivity(self):
        return self._compute_absorptance()

    def _compute_absorptance(self):
        a = Bandpass.zero()
        for i, el in enumerate(self._elements[:-1]):
            partial_absorptance = \
                el['element'].absorptance * self._compute_transmittance(
                    from_element=i + 1)
            a = self._sum_absorptance(a, partial_absorptance)
        a = self._sum_absorptance(a, self._elements[-1]['element'].absorptance)
        return a

    def _sum_absorptance(self, a, b):
        return Bandpass.from_array(self._waveset,
                                   a(self._waveset) + b(self._waveset))

    @property
    def elements(self):
        return self._elements

    def element_idx_from_name(self, name):
        el_idx = np.argwhere(
            list(map(lambda x: x['name'] == name, self._elements)))
        return el_idx[0][0]

    def element_from_name(self, name):
        return self._elements[self.element_idx_from_name(name)]['element']

    def as_transmissive_element(self):
        return TransmissiveElement(transmittance=self.transmittance,
                                   absorptance=self.emissivity)

    def plot(self, **kwargs):
        self.as_transmissive_element().plot(reflectance=False, **kwargs)

    def list_elements_transmittance(self, band):
        """
        Print the transmittance (or reflectance) in band for each element of the system.
        band: tuple (min, max) with band limits (units compatible with waveset)
        """
        for el in self._elements:
            name = el['name'] if el['name'] else '(no name)'
            id = el['element'].id
            direction = el['direction']
            elem = el['element']
            if direction == Direction.TRANSMISSION:
                if hasattr(elem, 'transmittance_in_band'):
                    val = elem.transmittance_in_band(band)
                    print(f"{name} ({id}) [T] {val[0]:.3f}")
                else:
                    print(f"{name} ({id}) [T] N/A")
            elif direction == Direction.REFLECTION:
                if hasattr(elem, 'reflectance_in_band'):
                    val = elem.reflectance_in_band(band)
                    print(f"{name} ({id}) [R] {val[0]:.3f}")
                else:
                    print(f"{name} ({id}) [R] N/A")
            else:
                print(f"{name} ({id}) ? N/A")
        total_val = self.total_transmittance(band)
        print(f"Total ({id}) [T] {total_val:.3f}")

    def print_elements_list(self):
        """
        Prints the list of elements with name and direction.
        """
        print(f"System: {self._name}")
        for i, el in enumerate(self._elements):
            name = el['name'] if el['name'] else '(no name)'
            id = el['element'].id
            if el['direction'] == Direction.REFLECTION:
                direction = 'R'
            elif el['direction'] == Direction.TRANSMISSION:
                direction = 'T'
            else:
                direction = '?'
            print(f"{i}: {name} ({id}) [{direction}]")
            
    def total_transmittance(self, band):
        """
        Print the total transmittance (or reflectance) in band for the system.
        band: tuple (min, max) with band limits (units compatible with waveset)
        """
        elem = self.as_transmissive_element()
        val = elem.transmittance_in_band(band) 
        return val[0]

    @staticmethod
    def combine(*transmissive_systems, name="Combined Transmissive System"):
        """
        Combine multiple TransmissiveSystems into a single system.
        
        Parameters
        ----------
        *transmissive_systems : TransmissiveSystem or list of TransmissiveSystem
            Variable number of TransmissiveSystem objects to combine, or a single list of systems.
        
        name : str, optional
            Name for the combined transmissive system. Default is "Combined Transmissive System".
        
        Returns
        -------
        TransmissiveSystem
            A new TransmissiveSystem containing all elements from the input systems.
        
        Examples
        --------
        >>> elt_ts = TransmissiveSystem("ELT")
        >>> mpo_ts = TransmissiveSystem("MPO")
        >>> # Both syntaxes work:
        >>> combined = TransmissiveSystem.combine(elt_ts, mpo_ts, name="ELT+MPO")
        >>> combined = TransmissiveSystem.combine([elt_ts, mpo_ts], name="ELT+MPO")
        """
        # Se il primo argomento è una lista/tupla, usala
        if len(transmissive_systems) == 1 and isinstance(transmissive_systems[0], (list, tuple)):
            systems = transmissive_systems[0]
        else:
            systems = transmissive_systems
        
        combined_system = TransmissiveSystem(name=name)
        for ts in systems:
            combined_system.add(ts)
        return combined_system



from functools import cached_property
import numpy as np
from arte.dataelab.base_data import BaseData

class Base2dMap(BaseData):
    """Two-dimensional map for spatial data representation.
    
    This class represents 2D spatial maps derived from time-series data,
    such as camera images, wavefront maps on telescope pupil, or deformable
    mirror shapes. It handles mapping between 1D data vectors and 2D spatial
    representations.
    
    The map defines which elements are valid (value=1) and which are not used
    (value=0), enabling proper reshaping of vectorized data into 2D images.
    
    Parameters
    ----------
    map2d : ndarray
        2D boolean or integer array where valid elements are 1 (or True)
        and unused elements are 0 (or False)
    
    Examples
    --------
    >>> # Create a circular pupil map
    >>> pupil = np.zeros((100, 100))
    >>> y, x = np.ogrid[-50:50, -50:50]
    >>> mask = x**2 + y**2 <= 40**2
    >>> pupil[mask] = 1
    >>> map2d = Base2dMap(pupil)
    >>> print(map2d.nvalid)  # Number of valid pixels
    
    >>> # Remap 1D vector data to 2D image
    >>> data_1d = np.random.randn(100, map2d.nvalid)  # 100 frames
    >>> data_2d = map2d.remap_image(data_1d)  # Shape: (100, 100, 100)
    """
    def __init__(self, map2d):
        super().__init__(map2d)

    @staticmethod
    def from_idx1d(shape, idx1d):
        f2d = np.zeros(shape, dtype=bool)
        f2d.flat[idx1d] = True
        return Base2dMap(f2d)

    @cached_property
    def nvalid(self):
        '''Number of valid elements'''
        nonzero = np.nonzero(self.get_data())
        return len(nonzero[0])

    def as_mask(self):
        ''''
        Return a mask suitable to be used in a masked array
        '''
        return (1 - self.get_data()).astype(bool)

    def remap_image(self, data):
        '''
        Remap a 2d time series data into a 3d cube, where each
        cube slice is a 2d image. If a 1d array is passed,
        the resulting 3d array will have a first dimension with length 1.
        
        Parameters
        ----------
        data: ndarray
            2d numpy array [time, values] or 1d numpy array [values]
        
        Returns
        -------
        frame: ndarray
            3d numpy array [time, x, y]
        '''
        if len(data.shape) == 1:
            data = np.expand_dims(data, axis=0)

        nonzero = np.nonzero(self.get_data())
        frame = np.zeros((len(data), self.shape[0], self.shape[1]), dtype=data.dtype)
        if len(nonzero[0]) != data.shape[1]:
            raise ValueError(f'Cannot build display: data shape and subap map shape do not match: %s and %d elements' % (data.shape, len(nonzero[0])))
        frame[:, nonzero[0], nonzero[1]] = data
        return frame



import numpy as np

from arte.types.mask import CircularMask
from arte.utils.decorator import override
from arte.optical_propagation.abstract_coronograph import Coronograph    


class FourQuadrantCoronograph(Coronograph):
    """ Class to simulate the 4-quadrant coronograph.
    The 4-quadrant coronograph applies a phase shift of
    pi to two opposite quadrants in the focal plane.

    The 4-quadrant coronograph is described in:
    "The four-quadrant phase-mask coronagraph: white light laboratory results with an achromatic device"
    D.  Mawet, P.  Riaud, J.  Baudrand, P.  Baudoz, A.  Boccaletti, O.  Dupuis, D.  Rouan
    A&A 448 (2) 801-808 (2006), DOI: 10.1051/0004-6361:20054158

    Parameters
    ----------
    referenceLambdaInM: float
        The wavelength at which the coronograph is defined [m].
        This is used to determine the focal plane mask size.
    
    outPupilStopInFractionOfPupil: float (optional)
        Outer radius of the pupil plane stop, in ratio of 
        pupil size (1 = full pupil, 0.3 = 30% pupil).
        By default 1, meaning that the pupil is not masked.

    inPupilStopInFractionOfPupil: float (optional)
        Inner radius of the pupil plane stop, in ratio of 
        pupil size (1 = full pupil, 0.3 = 30% pupil).
        By default 0, meaning that the pupil is not masked.

    Example
    -------
    fq = FourQuadrantCoronograph(referenceLambdaInM=800e-9,
                                 outPupilStopInFractionOfPupil=0.95)
    psf = fq.get_coronographic_psf(input_field, oversampling=4, lambdaInM=1000e-9)
    """

    def __init__(self,
                referenceLambdaInM:float,
                outPupilStopInFractionOfPupil:float=1.0,
                inPupilStopInFractionOfPupil:float=0.0):
        self._refLambdaInM = referenceLambdaInM
        self._inPupilStopSize = inPupilStopInFractionOfPupil
        self._outPupilStopSize = outPupilStopInFractionOfPupil

    @override
    def _get_pupil_mask(self, field):
        inStop = CircularMask(field.shape, maskRadius=self._inPupilStopSize*max(field.shape)/self.oversampling)
        outStop = CircularMask(field.shape, maskRadius=self._outPupilStopSize*max(field.shape)/self.oversampling)
        pupil_mask = np.logical_and(outStop.asTransmissionValue(),inStop.mask())
        return pupil_mask
    
    @override
    def _get_focal_plane_mask(self, field):
        nx,ny = field.shape
        cx,cy = nx//2,ny//2
        X,Y = np.mgrid[0:nx,0:ny]
        fmask = np.ones(field.shape)*np.pi
        top_left = np.logical_and(X>=cx, Y<cy)
        bottom_right = np.logical_and(X<cx, Y>=cy)
        fmask[top_left] = 0
        fmask[bottom_right] = 0
        if self.lambdaInM is not None:
            fmask *= self._refLambdaInM/self.lambdaInM
        focal_mask = np.exp(1j*fmask)
        return focal_mask



'''
Axis transposition handler for multi-dimensional time series data.

This module provides utilities for managing named axes in time series arrays
and performing axis reordering based on axis names rather than numeric indices.
'''

class AxisHandler():
    '''
    Handle axis transpositions in multi-dimensional arrays using named axes.
    
    This class enables intuitive axis reordering for time series data with
    multiple ensemble dimensions. Instead of using numeric indices, axes can
    be referenced by meaningful names (e.g., 'subaperture', 'wavelength', 'mode').
    
    Parameters
    ----------
    axes : sequence of str or str, optional
        Names for each axis in the data (excluding time axis, which is always first).
        If None, no axis transposition will be possible.
        If a single string, wraps it in a tuple.
    
    Examples
    --------
    >>> # Define 3D time series: (time, subaperture, wavelength)
    >>> handler = AxisHandler(axes=['subaperture', 'wavelength'])
    >>> 
    >>> # Transpose to (subaperture, wavelength, time)
    >>> data_transposed = handler.transpose(data, ['subaperture', 'wavelength'])
    >>> 
    >>> # Reorder axes: (time, wavelength, subaperture)
    >>> data_reordered = handler.transpose(data, ['wavelength', 'subaperture'])
    '''

    def __init__(self, axes=None):
        '''
        Initialize axis handler.

        Parameters
        ----------
        axes : sequence of str or str, optional
            Names for data axes. If None, no transposition possible.
        '''
        if axes is None:
            self._axes = ()
        elif isinstance(axes, str):
            self._axes = (axes,)
        else:
            self._axes = tuple(axes)

    def transpose(self, data, axes=None):
        '''
        Transpose array axes according to named axis order.
        
        Parameters
        ----------
        data : ndarray
            Array to transpose. First dimension (time) is not affected.
        axes : sequence of str or str, optional
            Desired axis order using axis names defined during initialization.
            If None, no transposition is performed.
        
        Returns
        -------
        ndarray
            Transposed array with axes reordered as specified.
        
        Raises
        ------
        ValueError
            If any requested axis name was not defined during initialization.
        
        Examples
        --------
        >>> handler = AxisHandler(axes=['x', 'y', 'wavelength'])
        >>> data = np.random.rand(100, 10, 20, 5)  # (time, x, y, wavelength)
        >>> 
        >>> # Reorder to (time, wavelength, x, y)
        >>> data_new = handler.transpose(data, ['wavelength', 'x', 'y'])
        >>> data_new.shape  # (100, 5, 10, 20)
        '''
        if axes is not None:
            if type(axes) is str:
                axes = [axes]
            # Use an explicit for loop in order to capture an erroneus axis name
            idx = []
            try:
                for ax in axes:
                    idx.append(self._axes.index(ax))
            except ValueError:
                raise ValueError(f'Axis {ax} not found')
            data = data.transpose(idx)
        return data

    def axes(self):
        return self._axes



import abc
import math
import collections
from functools import cache, cached_property

import numpy as np
from scipy.signal import welch
from astropy import units as u

from arte.time_series.axis_handler import AxisHandler
from arte.utils.not_available import NotAvailable
from arte.utils.help import add_help, modify_help
from arte.utils.unit_checker import make_sure_its_a


class TimeSeriesException(Exception):
    '''Exception raised by TimeSeries'''


@add_help
class TimeSeries(metaclass=abc.ABCMeta):
    '''
    Base class for time series analysis of adaptive optics telemetry data.

    This class provides a framework for analyzing time-evolving data where multiple
    "sister" quantities (ensemble elements) are sampled simultaneously over time.
    Examples include:
    
    - Camera pixel intensities evolving over multiple frames
    - Deformable mirror actuator commands over time
    - Wavefront sensor slopes for multiple subapertures
    - Modal coefficients (Zernike, KL modes) during AO operation
    
    The data structure is always `(n_time_samples, ...ensemble_shape...)` where
    the first dimension is time and remaining dimensions define the ensemble geometry.
    
    Subclasses must implement:
    
    - :meth:`_get_not_indexed_data`: Return raw data as numpy array with time as first axis
    - :meth:`get_index_of`: Define ensemble indexing logic (select specific elements)
    
    The class provides a **chainable fluent API** for data analysis:
    
    - Ensemble reductions: `ensemble_rms`, `ensemble_mean`, `ensemble_std`, `ensemble_median`, `ensemble_ptp`
    - Time reductions: `time_mean`, `time_std`, `time_rms`, `time_median`, `time_ptp`
    - Filtering: `filter(modes=..., times=...)` or `with_times(...)`
    - Value extraction: `.value` property (like pandas)
    - Chain operations: `ts.filter(modes=[2,3]).ensemble_rms.time_mean.value`
    
    Additional features:
    
    - Power spectral density analysis with Welch method
    - Physical units support via astropy.units
    - Masked arrays for missing/invalid data (e.g., points outside circular pupil)
    
    Parameters
    ----------
    axes : sequence of str, optional
        Named axes for the ensemble dimensions, enabling axis transposition.
        If None, no axis reordering is possible.
    
    Notes
    -----
    MaskedArray support: Use `numpy.ma.MaskedArray` to handle missing data in the
    ensemble dimension (e.g., wavefront defined only inside a circular pupil).
    Temporal masking is possible but may cause issues with FFT-based operations
    if sampling becomes non-uniform.
    
    Originally implemented as part of the ARGOS codebase.
    
    See Also
    --------
    MultiTimeSeries : Combine multiple TimeSeries with different sampling rates
    
    Examples
    --------
    >>> class MyTimeSeries(TimeSeries):
    ...     def __init__(self, data):
    ...         super().__init__()
    ...         self._data = data  # shape: (n_times, n_elements)
    ...     
    ...     def _get_not_indexed_data(self):
    ...         return self._data
    ...     
    ...     def get_index_of(self, *args, **kwargs):
    ...         if len(args) == 0:
    ...             return None  # Return all elements
    ...         return args[0]   # Return specific index
    >>> 
    >>> # Analyze DM commands for 1000 time steps, 100 actuators
    >>> dm_commands = np.random.randn(1000, 100)
    >>> ts = MyTimeSeries(dm_commands)
    >>> 
    >>> # Fluent API (chainable)
    >>> mean_rms = ts.ensemble_rms.time_mean.value  # Scalar
    >>> rms_time_series = ts.filter(elements=[0,1,2]).ensemble_rms.value  # Array
    >>> long_exposure_ptp = ts.time_mean.ensemble_ptp.value  # Peak-to-valley
    '''

    def __init__(self, axes=None):
        self._frequency = None
        self._last_cut_frequency = None
        self._power = None
        self._segment_factor = None
        self._window = None
        self._axis_handler = AxisHandler(axes)

    def __array__(self, dtype=None, copy=None):
        """
        Numpy compatibility protocol (numpy 2.0+ compatible).
        
        Enables using TimeSeries directly in numpy operations and matplotlib:
        - plt.plot(time, ts) - automatic conversion
        - np.mean(ts) - works directly
        
        Parameters
        ----------
        dtype : numpy dtype, optional
            Desired dtype for the array (numpy 2.0+)
        copy : bool, optional
            Whether to copy the data (numpy 2.0+)
        
        Returns
        -------
        ndarray
            Underlying data array
        
        Examples
        --------
        >>> ts = MyTimeSeries(data)
        >>> rms = ts.ensemble_rms  # Returns TimeSeries
        >>> plt.plot(rms)  # Works! __array__() called automatically
        >>> mean_rms = np.mean(rms)  # Works!
        """
        arr = self._get_not_indexed_data()
        if dtype is not None:
            arr = arr.astype(dtype)
        if copy:
            arr = arr.copy()
        return arr
    
    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """
        Support numpy universal functions on TimeSeries.
        
        Enables operations like: np.sin(ts), np.sqrt(ts), ts * 2
        
        Parameters
        ----------
        ufunc : numpy ufunc
            The universal function being applied
        method : str
            The ufunc method ('__call__', 'reduce', etc.)
        *inputs : tuple
            Input arrays/objects
        **kwargs : dict
            Additional ufunc arguments
        
        Returns
        -------
        ndarray or scalar
            Result of the ufunc operation
        """
        # Convert TimeSeries inputs to arrays
        arrays = []
        for inp in inputs:
            if isinstance(inp, TimeSeries):
                arrays.append(inp.__array__())
            else:
                arrays.append(inp)
        
        # Apply ufunc and return result (not wrapped in TimeSeries)
        return getattr(ufunc, method)(*arrays, **kwargs)

    @abc.abstractmethod
    def _get_not_indexed_data(self):
        pass

    @property
    def axes(self):
        '''Data series shape'''
        return self._axis_handler.axes()

    @cache
    def _get_time_vector(self):
        '''Override to provide a custom time vector'''
        return np.arange(len(self._get_not_indexed_data()))
    
    def _get_time_vector_check(self):
        v = self._get_time_vector()
        if isinstance(v, np.ndarray):
            return v
        else:
            try:
                # Try to convert
                return np.array(v)
            except TypeError as e:
                raise TimeSeriesException('Cannot convert _get_time_vector() result to numpy array') from e

    def get_data(self, *args, times=None, axes=None, **kwargs):
        '''
        Retrieve time series data with optional filtering and indexing.
        
        Parameters
        ----------
        *args, **kwargs
            Passed to :meth:`get_index_of` for ensemble element selection.
        times : sequence of 2 elements, optional
            Time range `[start, stop]` for filtering. Can be None (no filtering),
            or contain None elements for one-sided filtering.
            Units must match the time vector if using astropy quantities.
        axes : sequence of str, optional
            Reorder axes to specified order. Requires axis names to be defined
            during initialization.
        
        Returns
        -------
        ndarray
            Data array with shape `(n_times, ...ensemble_shape...)` after
            applying time filtering, ensemble indexing, and axis transposition.
        
        Examples
        --------
        >>> # Get all data
        >>> data = ts.get_data()
        >>> 
        >>> # Get data for specific time range (closed-loop only)
        >>> data = ts.get_data(times=[1.0 * u.s, 5.0 * u.s])
        >>> 
        >>> # Get data for specific ensemble elements
        >>> data = ts.get_data(modes=[2, 3, 4])  # Assuming mode-based indexing
        '''

        data = self._get_not_indexed_data()
        if times is not None:
            if not isinstance(times, collections.abc.Sequence) or len(times) != 2:
                raise TimeSeriesException('Times keywords must be a sequence of two elements: [start, stop]')
            time_vector = self.get_time_vector()
            if len(time_vector) != len(data):
                raise TimeSeriesException('Time vector and data lengths differ')

            start, stop = times
            if isinstance(time_vector, u.Quantity):
                if start is not None:
                    start = make_sure_its_a(time_vector.unit, start)
                if stop is not None:
                    stop = make_sure_its_a(time_vector.unit, stop)

            idxs = np.ones(len(time_vector), dtype=bool)
            if start is not None:
                idxs = np.logical_and(idxs, time_vector >= start)
            if stop is not None:
                idxs = np.logical_and(idxs, time_vector < stop)
            data = data[idxs]

        index = self.get_index_of(*args, **kwargs)
        data = self._index_data(data, index)
        data = self._axis_handler.transpose(data, axes)
        return data

    def _index_data(self, data, index):
        if index is None:
            return data
        elif isinstance(index, tuple):
            return data[(slice(0,data.shape[0]),) + index]
        else:
            return data[:, index]

    @abc.abstractmethod
    def get_index_of(self, *args, **kwargs):
        pass

    def data_unit(self):
        '''Override to return a string with a compact unit notation'''
        return None

    def data_label(self):
        '''Override to return a string with a readable unit name'''
        return None

    @cached_property
    def delta_time(self):
        '''Property with the interval between samples.
        
        If no interval can be determined (time vector too short),
        returns 1 with the correct unit if applicable.
        '''
        time_vector = self.get_time_vector()
        if len(time_vector) == 0:
            return 1
        diff = np.diff(time_vector)
        if len(diff) > 0:
            return np.median(diff)
        else:
            if isinstance(time_vector[0], u.Quantity):
                return 1 * time_vector[0].unit
            else:
                return 1

    def frequency(self):
        return self._frequency

    def get_time_vector(self):
        '''Return the series time vector'''
        return self._get_time_vector_check()

    def last_cut_frequency(self):
        return self._last_cut_frequency

    @cache
    def ensemble_size(self):
        '''Number of distinct series in this time ensemble'''
        not_indexed_data = self._get_not_indexed_data()
        return math.prod(not_indexed_data.shape[1:])

    @cache
    def time_size(self):
        '''Number of time samples in this time ensemble'''
        not_indexed_data = self._get_not_indexed_data()
        return not_indexed_data.shape[0]

    def _data_flattened(self, data):
        '''Flatten data over all non-time dimensions'''
        return data.reshape(data.shape[0], self._data_sample_size(data))

    def _data_sample_shape(self, data):
        return data.shape[1:]

    def _data_sample_axes(self, data):
        return tuple(range(len(data.shape))[1:])

    def _data_sample_size(self, data):
        return math.prod(data.shape[1:])

    @modify_help(arg_str='[series_idx], [times=[from,to]]')
    def get_time_median(self, *args, times=None, **kwargs):
        '''Median over time for each series
        
        .. deprecated::
           Use the `time_median` property instead. This method will be removed in a future version.
        '''
        import warnings
        warnings.warn(
            "get_time_median() is deprecated. Use the 'time_median' property instead.",
            DeprecationWarning,
            stacklevel=2
        )
        data = self.get_data(*args, times=times, **kwargs)
        if isinstance(data, np.ma.MaskedArray):
            return np.ma.median(data, axis=0)
        return np.median(data, axis=0)

    @modify_help(arg_str='[series_idx], [times=[from,to]]')
    def get_time_std(self, *args, times=None, **kwargs):
        '''Standard deviation over time for each series
        
        .. deprecated:: 
           Use the `time_std` property instead. This method will be removed in a future version.
        '''
        import warnings
        warnings.warn(
            "get_time_std() is deprecated. Use the 'time_std' property instead.",
            DeprecationWarning,
            stacklevel=2
        )
        data = self.get_data(*args, times=times, **kwargs)
        if isinstance(data, np.ma.MaskedArray):
            return np.ma.std(data, axis=0)
        return np.std(data, axis=0)

    @modify_help(arg_str='[series_idx], [times=[from,to]]')
    def get_time_average(self, *args, times=None, **kwargs):
        '''Average value over time for each series
        
        .. deprecated:: 
           Use the `time_mean` property instead. This method will be removed in a future version.
        '''
        import warnings
        warnings.warn(
            "get_time_average() is deprecated. Use the 'time_mean' property instead.",
            DeprecationWarning,
            stacklevel=2
        )
        data = self.get_data(*args, times=times, **kwargs)
        if isinstance(data, np.ma.MaskedArray):
            return np.ma.mean(data, axis=0)
        return np.mean(data, axis=0)

    @modify_help(arg_str='[series_idx], [times=[from,to]]')
    def get_time_rms(self, *args, times=None, **kwargs):
        '''Root-Mean-Square value over time for each series
        
        .. deprecated::
           Use the `time_rms` property instead. This method will be removed in a future version.
        '''
        import warnings
        warnings.warn(
            "get_time_rms() is deprecated. Use the 'time_rms' property instead.",
            DeprecationWarning,
            stacklevel=2
        )
        x = self.get_data(*args, times=times, **kwargs)
        if isinstance(x, np.ma.MaskedArray):
            return np.sqrt(np.ma.mean(np.abs(x)**2, axis=0))
        return np.sqrt(np.mean(np.abs(x)**2, axis=0))

    @modify_help(arg_str='[series_idx], [times=[from,to]]')
    def get_time_ptp(self, *args, times=None, **kwargs):
        '''Peak-to-peak (max - min) over time for each series
        
        .. deprecated::
           Use the `time_ptp` property instead. This method will be removed in a future version.
        '''
        import warnings
        warnings.warn(
            "get_time_ptp() is deprecated. Use the 'time_ptp' property instead.",
            DeprecationWarning,
            stacklevel=2
        )
        data = self.get_data(*args, times=times, **kwargs)
        if isinstance(data, np.ma.MaskedArray):
            return np.ma.ptp(data, axis=0)
        return np.ptp(data, axis=0)


    @modify_help(arg_str='[series_idx], [times=[from,to]]')
    def get_ensemble_average(self, *args, times=None, **kwargs):
        '''Average across series at each sampling time
        
        .. deprecated::
           Use the `ensemble_mean` property instead. This method will be removed in a future version.
        '''
        import warnings
        warnings.warn(
            "get_ensemble_average() is deprecated. Use the 'ensemble_mean' property instead.",
            DeprecationWarning,
            stacklevel=2
        )
        data = self.get_data(*args, times=times, **kwargs)
        if isinstance(data, np.ma.MaskedArray):
            return np.ma.mean(data, axis=self._data_sample_axes(data))
        return np.mean(data, axis=self._data_sample_axes(data))

    @modify_help(arg_str='[series_idx], [times=[from,to]]')
    def get_ensemble_std(self, *args, times=None, **kwargs):
        '''Standard deviation across series at each sampling time
        
        .. deprecated::
           Use the `ensemble_std` property instead. This method will be removed in a future version.
        '''
        import warnings
        warnings.warn(
            "get_ensemble_std() is deprecated. Use the 'ensemble_std' property instead.",
            DeprecationWarning,
            stacklevel=2
        )
        data = self.get_data(*args, times=times, **kwargs)
        if isinstance(data, np.ma.MaskedArray):
            return np.ma.std(data, axis=self._data_sample_axes(data))
        return np.std(data, axis=self._data_sample_axes(data))

    @modify_help(arg_str='[series_idx], [times=[from,to]]')
    def get_ensemble_median(self, *args, times=None, **kwargs):
        '''Median across series at each sampling time
        
        .. deprecated::
           Use the `ensemble_median` property instead. This method will be removed in a future version.
        '''
        import warnings
        warnings.warn(
            "get_ensemble_median() is deprecated. Use the 'ensemble_median' property instead.",
            DeprecationWarning,
            stacklevel=2
        )
        data = self.get_data(*args, times=times, **kwargs)
        if isinstance(data, np.ma.MaskedArray):
            return np.ma.median(data, axis=self._data_sample_axes(data))
        return np.median(data, axis=self._data_sample_axes(data))

    @modify_help(arg_str='[series_idx], [times=[from,to]]')
    def get_ensemble_rms(self, *args, times=None, **kwargs):
        '''
        Root-Mean-Square across series at each sampling time (legacy method).
        
        .. deprecated:: 
            Use the chainable property :attr:`ensemble_rms` for fluent API:
            ``ts.ensemble_rms`` or ``ts.filter(modes=[2,3,4]).ensemble_rms``
            This method will be removed in a future version.
        
        Returns
        -------
        ndarray
            RMS values (not chainable)
        '''
        import warnings
        warnings.warn(
            "get_ensemble_rms() is deprecated. Use the chainable property 'ensemble_rms' instead: "
            "ts.ensemble_rms or ts.filter(modes=[2,3,4]).ensemble_rms",
            DeprecationWarning,
            stacklevel=2
        )
        data = self.get_data(*args, times=times, **kwargs)
        if isinstance(data, np.ma.MaskedArray):
            return np.sqrt(np.ma.mean(np.abs(data)**2, axis=self._data_sample_axes(data)))
        return np.sqrt(np.mean(np.abs(data)**2, axis=self._data_sample_axes(data)))

    @modify_help(arg_str='[series_idx], [times=[from,to]]')
    def get_ensemble_ptp(self, *args, times=None, **kwargs):
        '''Peak-to-peak (max - min) across series at each sampling time
        
        .. deprecated::
           Use the `ensemble_ptp` property instead. This method will be removed in a future version.
        '''
        import warnings
        warnings.warn(
            "get_ensemble_ptp() is deprecated. Use the 'ensemble_ptp' property instead.",
            DeprecationWarning,
            stacklevel=2
        )
        data = self.get_data(*args, times=times, **kwargs)
        if isinstance(data, np.ma.MaskedArray):
            return np.ma.ptp(data, axis=self._data_sample_axes(data))
        return np.ptp(data, axis=self._data_sample_axes(data))

    # --- Chainable (Fluent) API ---
    # Upstream filtering methods for fluent API
    
    def filter(self, *args, **kwargs):
        """
        Create filtered TimeSeries by applying ensemble and/or time selection.
        
        Accepts same arguments as get_data(), creating a new TimeSeries
        with pre-filtered data for fluent chaining. Works with any Indexer
        implementation (ModeIndexer, RowColIndexer, DefaultIndexer, custom).
        
        Parameters
        ----------
        *args, **kwargs
            Passed to get_index_of() for ensemble filtering.
            Common kwargs depend on the Indexer:
            - modes= : for ModeIndexer (mode numbers)
            - elements= : for DefaultIndexer (element indices)
            - rows=, cols= : for RowColIndexer (spatial coordinates)
            - coord=, axis= : for interleaved/sequential xy layouts
        times : sequence of 2 elements, optional
            Time range [start, stop] for temporal filtering.
            Can contain None for one-sided filtering.
        
        Returns
        -------
        TimeSeries
            New TimeSeries with filtered data (chainable)
        
        Examples
        --------
        >>> # Modal selection (uses ModeIndexer in derived classes)
        >>> ts.filter(modes=[2, 3, 4]).ensemble_rms.time_mean
        >>> 
        >>> # Time + modal selection
        >>> ts.filter(modes=[2, 3, 4], times=[0.5, 1.0]).ensemble_rms.time_mean
        >>> 
        >>> # Element selection (DefaultIndexer)
        >>> ts.filter(elements=[0, 2, 4, 6]).time_std
        >>> 
        >>> # Row/col selection (RowColIndexer for 2D data)
        >>> ts.filter(rows=slice(0, 10), cols=[5, 10]).ensemble_rms
        >>> 
        >>> # Chainable (filters are accumulated)
        >>> ts.filter(modes=[2, 3, 4]).filter(times=[0.5, 1.0]).ensemble_rms.time_mean
        >>> 
        >>> # Custom indexer kwargs work automatically
        >>> ts.filter(coord='x').time_mean  # for interleaved xy data
        """
        return self._create_filtered_series(*args, **kwargs)
    
    def with_times(self, times):
        """
        Filter to specific time interval (convenience alias for filter(times=...)).
        
        Returns a new chainable TimeSeries containing only data within
        the specified time range [start, stop].
        
        Note: This is an alias for filter(times=...). For filtering by both
        time and ensemble elements, use filter() directly.
        
        Parameters
        ----------
        times : sequence of 2 elements
            Time range [start, stop] (can contain None for one-sided filtering)
        
        Returns
        -------
        TimeSeries
            Time-filtered TimeSeries (chainable)
        
        Examples
        --------
        >>> # Filter time range and compute statistics
        >>> ts.with_times([1, 2]).ensemble_rms.time_mean
        >>> 
        >>> # Chain with filter() for ensemble selection
        >>> ts.filter(modes=[2, 3, 4]).with_times([1, 2]).ensemble_rms
        >>> 
        >>> # Equivalent using filter() directly (preferred)
        >>> ts.filter(modes=[2, 3, 4], times=[1, 2]).ensemble_rms
        """
        return self.filter(times=times)
    
    def _create_filtered_series(self, *args, **kwargs):
        """
        Create filtered view of this TimeSeries.
        
        Returns a new TimeSeries that lazily applies filtering when data
        is accessed. The filtered series supports further chaining.
        
        Parameters
        ----------
        *args
            Positional arguments passed to get_index_of()
        **kwargs
            Keyword arguments for filtering:
            - times : time range [start, stop]
            - Other kwargs passed to get_index_of() (modes=, elements=, rows=, cols=, etc.)
        
        Returns
        -------
        TimeSeries
            Filtered TimeSeries (chainable)
        """
        parent_ts = self
        
        # Extract times separately (handled specially by get_data)
        filter_times = kwargs.pop('times', None)
        filter_args = args
        filter_kwargs = kwargs
        
        class FilteredTimeSeries(TimeSeries):
            def __init__(self):
                super().__init__(axes=parent_ts.axes)
                self._parent = parent_ts
                self._filter_args = filter_args
                self._filter_kwargs = filter_kwargs
                self._filter_times = filter_times
            
            def _get_not_indexed_data(self):
                # Apply parent's filtering: delegate all to get_data()
                return self._parent.get_data(
                    *self._filter_args,
                    times=self._filter_times,
                    **self._filter_kwargs
                )
            
            def get_index_of(self, *args, **kwargs):
                # Delegate to parent's indexing
                return self._parent.get_index_of(*args, **kwargs)
            
            def _create_filtered_series(self, *args, **kwargs):
                # Support chaining: accumulate filters
                # Merge new filters with existing ones
                new_times = kwargs.pop('times', None)
                if new_times is None:
                    new_times = self._filter_times
                
                # Merge kwargs (new ones override existing)
                merged_kwargs = {**self._filter_kwargs, **kwargs}
                
                # For args, we can't easily merge, so new args override
                merged_args = args if args else self._filter_args
                
                return self._parent._create_filtered_series(
                    *merged_args,
                    times=new_times,
                    **merged_kwargs
                )
            
            @cache
            def _get_time_vector(self):
                # Time vector reflects filtering
                if self._filter_times is not None:
                    # Validate times parameter (same as get_data)
                    if not isinstance(self._filter_times, collections.abc.Sequence) or len(self._filter_times) != 2:
                        raise TimeSeriesException('Times keywords must be a sequence of two elements: [start, stop]')
                    
                    parent_time = self._parent.get_time_vector()
                    
                    # Apply time filtering logic with unit conversion (same as get_data)
                    start, stop = self._filter_times
                    if isinstance(parent_time, u.Quantity):
                        if start is not None:
                            start = make_sure_its_a(parent_time.unit, start)
                        if stop is not None:
                            stop = make_sure_its_a(parent_time.unit, stop)
                    
                    idxs = np.ones(len(parent_time), dtype=bool)
                    if start is not None:
                        idxs = np.logical_and(idxs, parent_time >= start)
                    if stop is not None:
                        idxs = np.logical_and(idxs, parent_time < stop)
                    
                    return parent_time[idxs]
                else:
                    return self._parent.get_time_vector()
            
            @cached_property
            def delta_time(self):
                return self._parent.delta_time
        
        return FilteredTimeSeries()
    
    # Chainable operations (properties return TimeSeries)
    
    def _create_reduced_series(self, data, operation_name='reduced'):
        """
        Create new TimeSeries after reduction operation.
        
        Helper for chainable operations that reduce dimensions.
        Preserves time-related metadata in the new series.
        
        Parameters
        ----------
        data : ndarray
            Reduced data (e.g., 1D after ensemble aggregation)
        operation_name : str, optional
            Name of operation for tracking/debugging
        
        Returns
        -------
        TimeSeries
            New TimeSeries instance wrapping the reduced data (chainable)
        
        Notes
        -----
        This creates a minimal TimeSeries subclass that wraps the reduced array.
        Time axis is preserved (axis 0), allowing further temporal operations.
        """
        # Create anonymous subclass with the reduced data
        parent_ts = self
        
        class ReducedTimeSeries(TimeSeries):
            def __init__(self):
                super().__init__()
                self._reduced_data = data
                self._parent = parent_ts
                self._operation = operation_name
            
            def _get_not_indexed_data(self):
                return self._reduced_data
            
            def get_index_of(self, *args, **kwargs):
                # Reduced series has no ensemble indexing
                return None
            
            def __getitem__(self, key):
                """Support numpy-style indexing"""
                return self._reduced_data[key]
            
            @cache
            def _get_time_vector(self):
                # Inherit time vector from parent
                return self._parent.get_time_vector()
            
            @cached_property
            def delta_time(self):
                # Inherit delta_time from parent
                return self._parent.delta_time
        
        return ReducedTimeSeries()
    
    def _create_temporal_reduced_series(self, data, operation_name='time_reduced'):
        """
        Create TimeSeries after temporal reduction operation.
        
        Used for operations that collapse the time dimension (e.g., time_mean, time_std).
        Returns a TimeSeries with time_size=1, using mean time as the single timestamp.
        
        Parameters
        ----------
        data : ndarray
            Temporally reduced data (no time axis, or time_size=1)
        operation_name : str, optional
            Name of operation for tracking/debugging
        
        Returns
        -------
        TimeSeries
            New TimeSeries with shape (1, ...ensemble_shape...) (chainable)
        
        Notes
        -----
        The time_vector for the reduced series contains a single element:
        the mean time of the original series.
        """
        parent_ts = self
        
        # Ensure data has time axis with size 1
        if data.ndim == 0:
            # Scalar → (1,) - use atleast_1d to preserve Quantity
            reduced_data = np.atleast_1d(data)
        elif data.shape[0] != 1:
            # Add time axis: (ensemble,) → (1, ensemble...)
            reduced_data = np.expand_dims(data, axis=0)
        else:
            reduced_data = data
        
        class TemporalReducedTimeSeries(TimeSeries):
            def __init__(self):
                super().__init__(axes=parent_ts.axes)
                self._reduced_data = reduced_data
                self._parent = parent_ts
                self._operation = operation_name
            
            def _get_not_indexed_data(self):
                return self._reduced_data
            
            def get_index_of(self, *args, **kwargs):
                # Delegate to parent for ensemble indexing
                return self._parent.get_index_of(*args, **kwargs)
            
            def __getitem__(self, key):
                """Support numpy-style indexing"""
                return self._reduced_data[key]
            
            @cache
            def _get_time_vector(self):
                # Single-element time vector: mean of parent times
                parent_time = self._parent.get_time_vector()
                if isinstance(parent_time, u.Quantity):
                    return np.array([np.mean(parent_time.value)]) * parent_time.unit
                else:
                    return np.array([np.mean(parent_time)])
            
            @cached_property
            def delta_time(self):
                # Delta time not meaningful for single-timestamp series
                return 0 if not isinstance(self._parent.delta_time, u.Quantity) else 0 * self._parent.delta_time.unit
        
        return TemporalReducedTimeSeries()
    
    @property
    def shape(self):
        """
        Shape of the underlying data array.
        
        Returns the shape tuple of the TimeSeries data, similar to numpy arrays.
        This provides convenient access without requiring explicit conversion.
        
        Returns
        -------
        tuple of int
            Shape of the data array (time, ensemble...)
        
        Examples
        --------
        >>> ts.shape
        (1000, 10)  # 1000 time steps, 10 modes
        >>> ts.ensemble_rms.shape
        (1000,)  # Collapsed to time dimension only
        """
        return self._get_not_indexed_data().shape
    
    @property
    def value(self):
        """
        Extract final value(s) from TimeSeries (convenience accessor).
        
        Returns the underlying data as a numpy array or scalar,
        similar to pandas .values. Useful for getting final results
        from chained operations without explicit numpy conversion.
        
        Returns
        -------
        float, int, or ndarray
            - If shape is (1,): returns scalar
            - If shape is (1, n): returns 1D array of length n
            - Otherwise: returns the full data array
        
        Examples
        --------
        >>> # Extract scalar from fully reduced series
        >>> scalar = ts.ensemble_rms.time_mean.value  # → float
        >>> 
        >>> # Extract array from partially reduced series  
        >>> array = ts.time_mean.value  # → ndarray(ensemble_shape)
        >>> 
        >>> # Works like pandas .values
        >>> values = ts.value  # → full data array
        """
        data = self._get_not_indexed_data()
        
        # Remove trivial dimensions and extract scalar if possible
        squeezed = np.squeeze(data)
        
        if squeezed.ndim == 0:
            # Scalar (0D array) → Python scalar
            return squeezed.item()
        else:
            return squeezed
    
    @property
    def ensemble_rms(self):
        """
        RMS over ensemble dimensions (chainable property).
        
        Returns a TimeSeries containing the RMS computed across all
        ensemble (non-time) dimensions at each time step.
        The result can be chained with time aggregation operations.
        
        Returns
        -------
        TimeSeries
            Time series of ensemble RMS values (numpy-compatible via __array__)
        
        Notes
        -----
        This property:
        - Works on ALL data (no filtering)
        - Returns TimeSeries (chainable with .time_mean, .time_std, etc.)
        - Is numpy-compatible via __array__ protocol
        
        For filtered data, use upstream filtering:
        ``ts.filter(modes=[2,3,4]).with_times([1,2]).ensemble_rms``
        
        For legacy array return, use: :meth:`get_ensemble_rms` (deprecated)
        
        Examples
        --------
        >>> # Chainable operations
        >>> ts = WavefrontSeries(...)  # shape (n_frames, ny, nx)
        >>> rms_series = ts.ensemble_rms  # shape (n_frames,) - chainable!
        >>> mean_rms = rms_series.time_mean  # scalar - fluent API!
        >>> 
        >>> # Numpy/matplotlib compatible:
        >>> plt.plot(time, rms_series)  # Works via __array__()
        >>> np.mean(rms_series)  # Works!
        >>> 
        >>> # With filtering (upstream):
        >>> mean_rms = ts.filter(modes=[2,3,4]).ensemble_rms.time_mean
        """
        data = self._get_not_indexed_data()
        axis = self._data_sample_axes(data)
        
        if isinstance(data, np.ma.MaskedArray):
            rms_data = np.sqrt(np.ma.mean(np.abs(data)**2, axis=axis))
        else:
            rms_data = np.sqrt(np.mean(np.abs(data)**2, axis=axis))
        
        return self._create_reduced_series(rms_data, operation_name='ensemble_rms')
    
    @property
    def time_mean(self):
        """
        Mean over time dimension (chainable property).
        
        Computes the average across time (axis 0), returning a
        TimeSeries with time_size=1 containing the temporal mean.
        Further operations (like ensemble_rms, ensemble_ptp) can be chained.
        
        Returns
        -------
        TimeSeries
            TimeSeries with shape (1, ...ensemble_shape...) (chainable)
            Use .value to extract the final array/scalar.
        
        Notes
        -----
        Chainable operation: returns TimeSeries (not array).
        For convenient array extraction, use .value:
        ``ts.time_mean.value`` → ndarray or scalar
        
        Examples
        --------
        >>> ts = WavefrontSeries(...)  # shape (n_frames, ny, nx)
        >>> mean_ts = ts.time_mean  # TimeSeries(1, ny, nx) - chainable!
        >>> mean_wf = ts.time_mean.value  # ndarray(ny, nx) - extracted
        >>> 
        >>> # Chaining works both ways now:
        >>> rms_series = ts.ensemble_rms  # (n_frames,)
        >>> mean_rms = rms_series.time_mean.value  # scalar
        >>> 
        >>> # NEW: time-then-ensemble operations
        >>> ptp_map = ts.time_mean.ensemble_ptp.value  # peak-to-valley of long-exposure
        """
        data = self._get_not_indexed_data()
        if isinstance(data, np.ma.MaskedArray):
            mean_data = np.ma.mean(data, axis=0)
        else:
            mean_data = np.mean(data, axis=0)
        
        return self._create_temporal_reduced_series(mean_data, operation_name='time_mean')
    
    @property
    def time_std(self):
        """
        Standard deviation over time dimension (chainable property).
        
        Computes the std across time (axis 0), returning a
        TimeSeries with time_size=1 containing the temporal std.
        Further operations (like ensemble_rms, ensemble_ptp) can be chained.
        
        Returns
        -------
        TimeSeries
            TimeSeries with shape (1, ...ensemble_shape...) (chainable)
            Use .value to extract the final array/scalar.
        
        Notes
        -----
        Chainable operation: returns TimeSeries (not array).
        For convenient array extraction, use .value:
        ``ts.time_std.value`` → ndarray or scalar
        
        Examples
        --------
        >>> ts = WavefrontSeries(...)  # shape (n_frames, ny, nx)
        >>> std_ts = ts.time_std  # TimeSeries(1, ny, nx) - chainable!
        >>> std_wf = ts.time_std.value  # ndarray(ny, nx) - extracted
        >>> 
        >>> # Chaining:
        >>> rms_series = ts.ensemble_rms  # (n_frames,)
        >>> std_rms = rms_series.time_std.value  # scalar
        >>> 
        >>> # NEW: time-then-ensemble operations
        >>> ptp_std = ts.time_std.ensemble_ptp.value  # peak-to-peak temporal variability
        """
        data = self._get_not_indexed_data()
        if isinstance(data, np.ma.MaskedArray):
            std_data = np.ma.std(data, axis=0)
        else:
            std_data = np.std(data, axis=0)
        
        return self._create_temporal_reduced_series(std_data, operation_name='time_std')
    
    @property
    def ensemble_mean(self):
        """
        Mean over ensemble dimensions (chainable property).
        
        Returns
        -------
        TimeSeries
            Mean across ensemble at each time step (chainable)
        
        Examples
        --------
        >>> mean_series = ts.ensemble_mean.value  # Shape: (n_times,)
        >>> overall_mean = ts.ensemble_mean.time_mean.value  # Scalar
        """
        data = self._get_not_indexed_data()
        axis = self._data_sample_axes(data)
        if isinstance(data, np.ma.MaskedArray):
            mean_data = np.ma.mean(data, axis=axis)
        else:
            mean_data = np.mean(data, axis=axis)
        return self._create_reduced_series(mean_data, operation_name='ensemble_mean')
    
    @property
    def ensemble_std(self):
        """
        Standard deviation over ensemble dimensions (chainable property).
        
        Returns
        -------
        TimeSeries
            Std across ensemble at each time step (chainable)
        
        Examples
        --------
        >>> std_series = ts.ensemble_std.value
        >>> mean_std = ts.ensemble_std.time_mean.value
        """
        data = self._get_not_indexed_data()
        axis = self._data_sample_axes(data)
        if isinstance(data, np.ma.MaskedArray):
            std_data = np.ma.std(data, axis=axis)
        else:
            std_data = np.std(data, axis=axis)
        return self._create_reduced_series(std_data, operation_name='ensemble_std')
    
    @property
    def ensemble_median(self):
        """
        Median over ensemble dimensions (chainable property).
        
        Returns
        -------
        TimeSeries
            Median across ensemble at each time step (chainable)
        
        Examples
        --------
        >>> median_series = ts.ensemble_median.value
        >>> time_avg_median = ts.ensemble_median.time_mean.value
        """
        data = self._get_not_indexed_data()
        axis = self._data_sample_axes(data)
        if isinstance(data, np.ma.MaskedArray):
            median_data = np.ma.median(data, axis=axis)
        else:
            median_data = np.median(data, axis=axis)
        return self._create_reduced_series(median_data, operation_name='ensemble_median')
    
    @property
    def ensemble_ptp(self):
        """
        Peak-to-peak over ensemble dimensions (chainable property).
        
        Returns
        -------
        TimeSeries
            Peak-to-peak across ensemble at each time step (chainable)
        
        Examples
        --------
        >>> ptp_series = ts.ensemble_ptp.value
        >>> mean_ptp = ts.ensemble_ptp.time_mean.value
        """
        data = self._get_not_indexed_data()
        axis = self._data_sample_axes(data)
        if isinstance(data, np.ma.MaskedArray):
            ptp_data = np.ma.ptp(data, axis=axis)
        else:
            ptp_data = np.ptp(data, axis=axis)
        return self._create_reduced_series(ptp_data, operation_name='ensemble_ptp')
    
    @property
    def time_median(self):
        """
        Median over time dimension (chainable property).
        
        Returns
        -------
        TimeSeries
            TimeSeries with time_size=1 containing temporal median (chainable)
        
        Examples
        --------
        >>> median_wf = ts.time_median.value
        >>> rms_of_median = ts.time_median.ensemble_rms.value
        """
        data = self._get_not_indexed_data()
        if isinstance(data, np.ma.MaskedArray):
            median_data = np.ma.median(data, axis=0)
        else:
            median_data = np.median(data, axis=0)
        return self._create_temporal_reduced_series(median_data, operation_name='time_median')
    
    @property
    def time_rms(self):
        """
        RMS over time dimension (chainable property).
        
        Returns
        -------
        TimeSeries
            TimeSeries with time_size=1 containing temporal RMS (chainable)
        
        Examples
        --------
        >>> rms_wf = ts.time_rms.value
        >>> max_rms = ts.time_rms.ensemble_ptp.value
        """
        data = self._get_not_indexed_data()
        if isinstance(data, np.ma.MaskedArray):
            rms_data = np.sqrt(np.ma.mean(np.abs(data)**2, axis=0))
        else:
            rms_data = np.sqrt(np.mean(np.abs(data)**2, axis=0))
        return self._create_temporal_reduced_series(rms_data, operation_name='time_rms')
    
    @property
    def time_ptp(self):
        """
        Peak-to-peak over time dimension (chainable property).
        
        Returns
        -------
        TimeSeries
            TimeSeries with time_size=1 containing temporal peak-to-peak (chainable)
        
        Examples
        --------
        >>> ptp_wf = ts.time_ptp.value
        >>> mean_ptp = ts.time_ptp.ensemble_mean.value
        """
        data = self._get_not_indexed_data()
        if isinstance(data, np.ma.MaskedArray):
            ptp_data = np.ma.ptp(data, axis=0)
        else:
            ptp_data = np.ptp(data, axis=0)
        return self._create_temporal_reduced_series(ptp_data, operation_name='time_ptp')

    @modify_help(call='power(from_freq=xx, to_freq=xx, [series_idx])')
    def power(self, *args, from_freq=None, to_freq=None,
              segment_factor=None, window='boxcar', **kwargs):
        '''
        Compute Power Spectral Density using Welch's method.
        
        Parameters
        ----------
        *args, **kwargs
            Passed to :meth:`get_index_of` for ensemble element selection.
        from_freq : float, optional
            Lower frequency bound for output. If None, starts from DC.
        to_freq : float, optional
            Upper frequency bound for output. If None, goes to Nyquist.
        segment_factor : float, optional
            Segment length factor for Welch method. Larger values give better
            frequency resolution but worse variance. Default is 1.0 (one segment).
        window : str, optional
            Window function name for Welch method. Default is 'boxcar' (no windowing).
            See scipy.signal.welch for available windows.
        
        Returns
        -------
        ndarray
            Power spectral density with shape `(n_frequencies, ...ensemble_shape...)`.
            Units are data_unit^2 / Hz (standard PSD units) if applicable.
        
        Notes
        -----
        The PSD is computed using scipy.signal.welch in standard units [data_unit^2/Hz].
        Results are cached unless segment_factor or window changes.
        
        For MaskedArrays, masked values are filled with zeros before FFT computation.
        
        Examples
        --------
        >>> # Compute full PSD
        >>> psd = ts.power()
        >>> freq = ts.frequency()
        >>> 
        >>> # Analyze only low frequencies (< 100 Hz)
        >>> psd = ts.power(from_freq=0, to_freq=100)
        >>> 
        >>> # Use Hanning window with 4 segments for variance reduction
        >>> psd = ts.power(segment_factor=4, window='hann')
        '''

        if segment_factor is None:
            if self._segment_factor is None:
                self._segment_factor = 1.0
        else:
            if self._segment_factor != segment_factor:
                self._segment_factor = segment_factor
                self._power = None
        if self._window != window:
            self._power = None
            self._window = window
        if self._power is None:
            data = self._get_not_indexed_data()
            # Perform power on flattened data and restore the shape afterwards
            power = self._compute_power(self._data_flattened(data))
            self._power = power.reshape((power.shape[0],) + self._data_sample_shape(data))
        if from_freq is None:
            output = self._power
            self._last_cut_frequency = self._frequency
        else:
            ul = self._frequency <= to_freq
            dl = self._frequency >= from_freq
            ul |= np.isclose(self._frequency, to_freq)
            dl |= np.isclose(self._frequency, from_freq)
            lim = ul & dl
            self._last_cut_frequency = self._frequency[lim]
            output = self._power[lim]
        index = self.get_index_of(*args, **kwargs)
        return self._index_data(output, index)

    def _compute_power(self, data):

        if isinstance(data, NotAvailable):
            raise Exception('Cannot calculate power: data is not available')
        
        # Convert MaskedArray to regular array for scipy.signal.welch compatibility
        if isinstance(data, np.ma.MaskedArray):
            data = data.filled(0)
        
        delta_time = self.delta_time
        if isinstance(delta_time, u.Quantity):
            value_hz = (1 / delta_time).to_value(u.Hz)
        else:
            value_hz = 1 / delta_time

        self._frequency, x = welch(data.T, value_hz,
                                   window=self._window,
                                   nperseg=data.shape[0] / self._segment_factor)
        # Return PSD in standard units [data_unit^2/Hz] as computed by scipy.signal.welch
        return x.T


# ___oOo___



import astropy.units as u

from arte.dataelab.base_timeseries import BaseTimeSeries
from arte.time_series.indexer import Indexer
from arte.utils.unit_checker import separate_value_and_unit

signal_unit = u.def_unit('signal')



class BaseSlopes(BaseTimeSeries):
    """Wavefront sensor slopes time series.
    
    This class handles slopes (wavefront gradients) recorded from wavefront
    sensors. Slopes are stored as vectors where typically half the elements
    are x-slopes and half are y-slopes.
    
    By default, slopes are in xxxxyyyy order (all X slopes followed by all Y
    slopes). Individual slope components can be accessed using 'x' and 'y'
    selectors.
    
    Slopes can be remapped into 2D arrays representing the telescope pupil,
    where each subaperture position shows its measured slope value.
    
    Parameters
    ----------
    data : array_like or DataLoader
        Slope data with shape (nframes, nslopes) or (nslopes,)
    time_vector : array_like or DataLoader, optional
        Time vector for each frame
    astropy_unit : astropy.units.Unit, optional
        Physical unit for slope values (default: signal_unit)
    data_label : str, optional
        Label for plots (default: 'slopes')
    axes : sequence, optional
        Names for data axes
    
    Examples
    --------
    >>> slopes = BaseSlopes('slopes.fits')
    >>> sx = slopes.get_data('x')  # Get x-slopes only
    >>> sy = slopes.get_data('y')  # Get y-slopes only
    >>> slopes.imshow()  # Display as 2D pupil maps
    >>> slopes.vecshow()  # Display as vector field
    
    Notes
    -----
    Derived classes can customize the slope order by providing a different
    indexer implementation.
    """

    def __init__(self, data, time_vector=None, astropy_unit=signal_unit, data_label='slopes', axes=None):
        super().__init__(data=data,
                         time_vector=time_vector,
                         astropy_unit=astropy_unit,
                         data_label=data_label,
                         axes=axes)
        self._indexer = Indexer()

    def get_index_of(self, *args, **kwargs):
        return self._indexer.sequential_xy(self.ensemble_size(), *args, **kwargs)

    def imshow(self, cut_wings=0):
        '''
        Display X and Y slope 2d images
        cut_wings=x means that colorbar is saturated for array values below x percentile
        and above 100-x percentile. Default is 0, i.e. all data are displayed; values below
        0 are forced to 0, values above 50 are set to 50.
        '''
        title = "left:" + self._data_label + "-X, right:" + self._data_label + "-Y"
        super().imshow(cut_wings=cut_wings, title=title)

    def vecshow(self):
        '''Display slopes as vector field'''
        sx2d = self.get_display('x').mean(axis=0)
        sy2d = self.get_display('y').mean(axis=0)
        sx2d, _ = separate_value_and_unit(sx2d)
        sy2d, _ = separate_value_and_unit(sy2d)

        import matplotlib.pyplot as plt
        plt.quiver(sx2d, sy2d)
        return plt




import numpy as np
import astropy.units as u

from matplotlib import pyplot
from matplotlib.axes import Axes

from arte.utils.not_available import NotAvailable
from arte.utils.unit_checker import make_sure_its_a, separate_value_and_unit


def modalplot(residual_modes_vector, pol_modes_vector, unit=None,
            overplot=False, plot_to=None,
            title=None, xlabel='Mode index', ylabel='wavefront rms', add_unit_to_ylabel=True,
            res_label='residual modes', res_color='red',
            pol_label='POL modes', pol_color='black'):
    """Plot modal coefficients comparison for adaptive optics analysis.
    
    Creates a plot comparing residual modal coefficients against POL (or other
    reference) modes. Typically used to visualize loop performance by plotting
    closed-loop residuals vs. open-loop measurements.
    
    Parameters
    ----------
    residual_modes_vector : array_like or astropy.units.Quantity
        Residual mode coefficients (typically RMS or std dev per mode)
    pol_modes_vector : array_like or astropy.units.Quantity
        Reference mode coefficients (typically POL modes or open-loop)
    unit : astropy.units.Unit, optional
        Target unit for both vectors. If specified, both inputs will be
        converted to this unit
    overplot : bool, optional
        If True, add to existing plot. If False, clear plot first
        (default: False)
    plot_to : matplotlib.pyplot or matplotlib.axes.Axes, optional
        Matplotlib object to plot to. If None, uses pyplot (default: None)
    title : str, optional
        Plot title. Required when overplot=False
    xlabel : str, optional
        X-axis label (default: 'Mode index')
    ylabel : str, optional
        Y-axis label (default: 'wavefront rms')
    add_unit_to_ylabel : bool, optional
        If True, append unit to ylabel (default: True)
    res_label : str, optional
        Legend label for residual modes (default: 'residual modes')
    res_color : str, optional
        Color for residual modes plot (default: 'red')
    pol_label : str, optional
        Legend label for POL modes (default: 'POL modes')
    pol_color : str, optional
        Color for POL modes plot (default: 'black')
    
    Returns
    -------
    matplotlib.axes.Axes
        The axes object containing the plot
    
    Examples
    --------
    >>> # Basic usage
    >>> modalplot(residuals, pol_modes, title='AO Performance')
    
    >>> # Apply optical gain and specify units
    >>> modalplot(
    ...     analyzer.residual_modes.time_std() * 2,
    ...     analyzer.pol_modes.time_std() * 2,
    ...     unit=u.nm,
    ...     title='Performance with 2x optical gain'
    ... )
    
    >>> # Overplot multiple datasets
    >>> modalplot(res1, pol1, title='Comparison')
    >>> modalplot(res2, pol2, overplot=True,
    ...          res_label='residual modes 2', res_color='blue')
    
    Notes
    -----
    - Automatically handles astropy units and NotAvailable values
    - When overplotting, title parameter is ignored
    - Legend is automatically generated from labels
    """

    if plot_to is not None:
        plt = plot_to
    else:
        plt = pyplot

    res_data = residual_modes_vector
    pol_data = pol_modes_vector

    if not overplot:
        plt.cla()
        plt.clf()
        if title is None:
            raise ValueError('Title must be set when not overplotting')

    # Make sure that the two units match if at all possible,
    # taking care of leaving NotAvailable values untouched.
    res_valid = not isinstance(res_data, NotAvailable)
    pol_valid = not isinstance(pol_data, NotAvailable)
    res_quantity = isinstance(res_data, u.Quantity)
    pol_quantity = isinstance(pol_data, u.Quantity)
    if unit is not None:
        res_data = make_sure_its_a(unit, res_data)
        pol_data = make_sure_its_a(unit, pol_data)
    elif res_valid and pol_valid and (res_quantity or pol_quantity):
        pol_data = make_sure_its_a(res_data.unit, pol_data)

    res_value, res_unit = separate_value_and_unit(res_data)
    pol_value, pol_unit = separate_value_and_unit(pol_data)

    # Select what to display on the Y axis.
    plot_unit = ''
    if res_quantity:
        plot_unit = res_unit
    elif pol_quantity:
        plot_unit = pol_unit
    if not isinstance(plot_unit, u.UnitBase):
        plot_unit = 'a.u.'
    
    if len(res_data) > 0:
        plt.plot(np.arange(len(res_data)), res_value,
                    color=res_color, label=res_label)
    if len(pol_data) > 0:
        plt.plot(np.arange(len(pol_data)), pol_value,
                    color=pol_color, label=pol_label)

    maxlen = max(len(res_data), len(pol_data))

    if add_unit_to_ylabel:
        ystr = f'{ylabel} [{plot_unit}]'
    else:
        ystr = ylabel

    if not overplot:
        if isinstance(plt, Axes):
            plt.set(xlabel=xlabel, ylabel=ystr)
            if title is not None:
                plt.set(title=title)
        else:
            plt.loglog()
            plt.xscale('symlog')    # Allow 0 index in log plots
#            plt.yscale('symlog')
            plt.xlim([0, maxlen])

            plt.xlabel(xlabel)
            plt.ylabel(ystr)
            if title is not None:
                plt.title(title)
            if not (res_label is None and pol_label is None):
                plt.legend()
    return plt



import numpy as np

from arte.types.mask import CircularMask
from arte.utils.decorator import override
from arte.optical_propagation.abstract_coronograph import Coronograph    


class LyotCoronograph(Coronograph):
    """ Class to simulate the Lyot coronograph 
    
    Parameters
    ----------
    referenceLambdaInM: float
        The wavelength at which the coronograph is defined [m].
        This is used to determine the focal plane mask size.

    inFocalStopInLambdaOverD: float
        Inner radius of the focal plane stop, in lambda/D units.

    outFocalStopInLambdaOverD: float (optional)
        Outer radius of the focal plane stop, in lambda/D units.
        By default None, meaning that higher spatial frequencies
        of the PSF are not filtered by a focal plane mask.
    
    outPupilStopInFractionOfPupil: float (optional)
        Outer radius of the pupil plane stop, in ratio of 
        pupil size (1 = full pupil, 0.3 = 30% pupil).
        By default 1, meaning that the pupil is not masked.

    inPupilStopInFractionOfPupil: float (optional)
        Inner radius of the pupil plane stop, in ratio of 
        pupil size (1 = full pupil, 0.3 = 30% pupil).
        By default 0, meaning that the pupil is not masked.

    Example
    -------
    lyot = LyotCoronograph(referenceLambdaInM=800e-9,
                           nFocalStopInLambdaOverD=2.0
                           outPupilStopInFractionOfPupil=0.95)
    psf = lyot.get_coronographic_psf(input_field, oversampling=4, lambdaInM=1000e-9) 
    """

    def __init__(self,
                referenceLambdaInM:float,
                inFocalStopInLambdaOverD:float,
                outFocalStopInLambdaOverD:float=None,
                outPupilStopInFractionOfPupil:float=1.0,
                inPupilStopInFractionOfPupil:float=0.0):
        self._refLambdaInM = referenceLambdaInM
        self._iwaInLambdaOverD = inFocalStopInLambdaOverD
        self._owaInLambdaOverD = outFocalStopInLambdaOverD
        self._inPupilStopSize = inPupilStopInFractionOfPupil
        self._outPupilStopSize = outPupilStopInFractionOfPupil

    @override
    def _get_pupil_mask(self, field):
        inStop = CircularMask(field.shape, maskRadius=self._inPupilStopSize*max(field.shape)/self.oversampling)
        outStop = CircularMask(field.shape, maskRadius=self._outPupilStopSize*max(field.shape)/self.oversampling)
        pupil_mask = np.logical_and(outStop.asTransmissionValue(),inStop.mask())
        return pupil_mask
    
    @override
    def _get_focal_plane_mask(self, field):
        lambdaInM2Px = self.oversampling
        if self.lambdaInM is not None:
            lambdaInM2Px *= self._refLambdaInM/self.lambdaInM
        iwa = CircularMask(field.shape, maskRadius=self._iwaInLambdaOverD*lambdaInM2Px)
        if self._owaInLambdaOverD is not None:
            owa = CircularMask(field.shape, maskRadius=self._owaInLambdaOverD*lambdaInM2Px)
            focal_mask = np.logical_and(iwa.mask(),owa.asTransmissionValue())
        else:
            focal_mask = iwa.mask()
        return focal_mask
    

class KnifeEdgeCoronograph(Coronograph):
    """ Class to simulate the Lyot coronograph 
    with a knife-edge focal plane mask.

    Parameters
    ----------
    referenceLambdaInM: float
        The wavelength at which the coronograph is defined [m].
        This is used to determine the focal plane mask size.
    
    iwaFocalStopInLambdaOverD: float
        Inner radius of the focal plane stop, in lambda/D units.
    
    outPupilStopInFractionOfPupil: float (optional)
        Outer radius of the pupil plane stop, in ratio of 
        pupil size (1 = full pupil, 0.3 = 30% pupil).
        By default 1, meaning that the pupil is not masked.

    inPupilStopInFractionOfPupil: float (optional)
        Inner radius of the pupil plane stop, in ratio of 
        pupil size (1 = full pupil, 0.3 = 30% pupil).
        By default 0, meaning that the pupil is not masked.

    Example
    -------
    knife_edge = KnifeEdgeCoronograph(referenceLambdaInM=800e-9,
                                      iwaFocalStopInLambdaOverD=2.0,
                                      outPupilStopInFractionOfPupil=0.95)
    psf = knife_edge.get_coronographic_psf(input_field, oversampling=4, lambdaInM=1000e-9)
    """

    def __init__(self,
                referenceLambdaInM:float,
                iwaFocalStopInLambdaOverD:float,
                outPupilStopInFractionOfPupil:float=1.0,
                inPupilStopInFractionOfPupil:float=0.0):
        self._refLambdaInM = referenceLambdaInM
        self._edgeInLambdaOverD = iwaFocalStopInLambdaOverD
        self._inPupilStopSize = inPupilStopInFractionOfPupil
        self._outPupilStopSize = outPupilStopInFractionOfPupil

    @override
    def _get_pupil_mask(self, field):
        inStop = CircularMask(field.shape, maskRadius=self._inPupilStopSize*max(field.shape)/self.oversampling)
        outStop = CircularMask(field.shape, maskRadius=self._outPupilStopSize*max(field.shape)/self.oversampling)
        pupil_mask = np.logical_and(outStop.asTransmissionValue(),inStop.mask())
        return pupil_mask
    
    @override
    def _get_focal_plane_mask(self, field):
        lambdaInM2Px = self.oversampling
        if self.lambdaInM is not None:
            lambdaInM2Px *= self._refLambdaInM/self.lambdaInM
        _,X = np.mgrid[0:field.shape[0],0:field.shape[1]]
        edge = (field.shape[1]//2+self._edgeInLambdaOverD*lambdaInM2Px)
        focal_mask = np.ones(field.shape)
        focal_mask[X<=edge] = 0
        return focal_mask





import numpy as np
import astropy.units as u

from arte.utils.help import modify_help
from arte.time_series.time_series import TimeSeries

class MultiTimeSeries(TimeSeries):
    '''
    Join multiple TimeSeries objects with incompatible time sampling.
    
    After being initialized with a list of
    :class:`~arte.time_series.time_series.TimeSeries` objects, which can
    have different ensemble sizes and length, it behaves as if all the data
    was part of a single ensemble. 
    
    **Time operations** (like :attr:`time_mean`, :attr:`time_std`) work in any case.
    
    **Ensemble operations** (like :attr:`ensemble_mean`, :attr:`ensemble_std`) 
    only work if all series have the same sampling rate (``delta_time``), 
    and raise an Exception otherwise. Use :meth:`is_homogeneous` to check 
    compatibility before calling ensemble operations.
    
    Examples
    --------
    >>> # Create multi-series with compatible sampling
    >>> multi = MultiTimeSeries(series1_1khz, series2_1khz)
    >>> multi.is_homogeneous()  # True - same sampling rates
    True
    >>> mean = multi.ensemble_mean.value  # Works!
    >>> 
    >>> # Incompatible sampling raises exception
    >>> multi2 = MultiTimeSeries(series_1khz, series_500hz)
    >>> multi2.ensemble_mean  # Raises Exception - incompatible rates
    >>> 
    >>> # Time operations always work
    >>> time_avg = multi2.time_mean.value  # Always works
    
    Notes
    -----
    Chainable API: ensemble operations return TimeSeries objects that can
    be chained with time operations and vice versa.
    '''
    def __init__(self, *args):
        super().__init__()

        self._series = []
        for v in args:
            self.add_series(v)

    def add_series(self, series):
        '''
        Adds a new series to this MultiTimeSeries instance
        
        Parameters
        ----------
        series: :class:`~arte.time_series.time_series.TimeSeries` or :class:`~arte.time_series.time_series.TimeSeriesWithInterpolation` instance
           the series to be added        
        '''
        self._series.append(series)

    def _get_not_indexed_data(self):
        return np.hstack([v._get_not_indexed_data()
                          for v in self._series])
        
    def is_homogeneous(self, *args, **kwargs):
        '''
        Check if selected series have compatible sampling rates.
        
        Parameters
        ----------
        *args, **kwargs
            Passed to :meth:`get_index_of` for series selection.
        
        Returns
        -------
        bool
            True if all selected series have identical `delta_time`,
            False otherwise.
        
        Notes
        -----
        This check is necessary before calling ensemble-wise operations
        like :attr:`ensemble_mean` or :attr:`ensemble_std`, which
        require uniform time sampling across all series.
        
        Examples
        --------
        >>> multi = MultiTimeSeries(series_1khz, series_500hz)
        >>> multi.is_homogeneous()  # False - different sampling rates
        False
        >>> 
        >>> multi2 = MultiTimeSeries(series1_1khz, series2_1khz)
        >>> multi2.is_homogeneous()  # True - same sampling
        True
        '''
        dt = [x.value for x in self.delta_times(*args, **kwargs)]
        return len(set(dt)) == 1

    def _impersonateDeltaTime(self, *args, **kwargs):
        '''Assume the first delta time of the ones selected by the args'''
        dt = self.delta_times(*args, **kwargs)
        self.delta_time = dt[0]

    def ensemble_size(self):
        '''Returns the total ensemble size'''
        return sum([x.ensemble_size() for x in self._series])

    def delta_times(self, *args, **kwargs):
        '''Returns a vector of delta times'''

        # Known astropy bug (numpy < 1.17): units are lost when using hstack 
        # We remove them before stacking, and add them later
        dt = np.hstack( \
              [np.repeat(x.delta_time.to('s').value, x.ensemble_size())
               for x in self._series])

        dt = dt * u.s

        index = self.get_index_of(*args, **kwargs)
        if index is not None and len(index)>0:
            dt = dt[index]

        return dt

    # Legacy methods (deprecated - use properties instead)
    
    @modify_help(arg_str='[time_idx]')
    def get_ensemble_average(self, *args, times=None, **kwargs):
        '''
        Average across series at each sampling time (legacy method).
        
        .. deprecated::
           Use the `ensemble_mean` property instead. This method will be removed in a future version.
        
        Returns
        -------
        ndarray
            Mean values (not chainable)
        
        Raises
        ------
        Exception
            If series have incompatible sampling rates (not homogeneous)
        '''
        import warnings
        warnings.warn(
            "get_ensemble_average() is deprecated. Use the 'ensemble_mean' property instead.",
            DeprecationWarning,
            stacklevel=2
        )
        if self.is_homogeneous(*args, **kwargs):
            self._impersonateDeltaTime(*args, **kwargs) 
            return super().get_ensemble_average(*args, times=times, **kwargs)
        else:
            raise Exception('Data series cannot be combined')

    @modify_help(arg_str='[time_idx]')
    def get_ensemble_std(self, *args, times=None, **kwargs):
        '''
        Standard deviation across series at each sampling time (legacy method).
        
        .. deprecated::
           Use the `ensemble_std` property instead. This method will be removed in a future version.
        
        Returns
        -------
        ndarray
            Std values (not chainable)
        
        Raises
        ------
        Exception
            If series have incompatible sampling rates (not homogeneous)
        '''
        import warnings
        warnings.warn(
            "get_ensemble_std() is deprecated. Use the 'ensemble_std' property instead.",
            DeprecationWarning,
            stacklevel=2
        )
        if self.is_homogeneous(*args, **kwargs):
            self._impersonateDeltaTime(*args, **kwargs) 
            return super().get_ensemble_std(*args, times=times, **kwargs)
        else:
            raise Exception('Data series cannot be combined')

    @modify_help(arg_str='[time_idx]')
    def get_ensemble_median(self, *args, times=None, **kwargs):
        '''
        Median across series at each sampling time (legacy method).
        
        .. deprecated::
           Use the `ensemble_median` property instead. This method will be removed in a future version.
        
        Returns
        -------
        ndarray
            Median values (not chainable)
        
        Raises
        ------
        Exception
            If series have incompatible sampling rates (not homogeneous)
        '''
        import warnings
        warnings.warn(
            "get_ensemble_median() is deprecated. Use the 'ensemble_median' property instead.",
            DeprecationWarning,
            stacklevel=2
        )
        if self.is_homogeneous(*args, **kwargs):
            self._impersonateDeltaTime(*args, **kwargs) 
            return super().get_ensemble_median(*args, times=times, **kwargs)
        else:
            raise Exception('Data series cannot be combined')

    @modify_help(arg_str='[time_idx]')
    def get_ensemble_rms(self, *args, times=None, **kwargs):
        '''
        Root-Mean-Square across series at each sampling time (legacy method).
        
        .. deprecated::
           Use the `ensemble_rms` property instead. This method will be removed in a future version.
        
        Returns
        -------
        ndarray
            RMS values (not chainable)
        
        Raises
        ------
        Exception
            If series have incompatible sampling rates (not homogeneous)
        '''
        import warnings
        warnings.warn(
            "get_ensemble_rms() is deprecated. Use the 'ensemble_rms' property instead.",
            DeprecationWarning,
            stacklevel=2
        )
        if self.is_homogeneous(*args, **kwargs):
            self._impersonateDeltaTime(*args, **kwargs) 
            return super().get_ensemble_rms(*args, times=times, **kwargs)
        else:
            raise Exception('Data series cannot be combined')

    @modify_help(arg_str='[time_idx]')
    def get_ensemble_ptp(self, *args, times=None, **kwargs):
        '''
        Peak-to-peak (max - min) across series at each sampling time (legacy method).
        
        .. deprecated::
           Use the `ensemble_ptp` property instead. This method will be removed in a future version.
        
        Returns
        -------
        ndarray
            Peak-to-peak values (not chainable)
        
        Raises
        ------
        Exception
            If series have incompatible sampling rates (not homogeneous)
        '''
        import warnings
        warnings.warn(
            "get_ensemble_ptp() is deprecated. Use the 'ensemble_ptp' property instead.",
            DeprecationWarning,
            stacklevel=2
        )
        if self.is_homogeneous(*args, **kwargs):
            self._impersonateDeltaTime(*args, **kwargs) 
            return super().get_ensemble_ptp(*args, times=times, **kwargs)
        else:
            raise Exception('Data series cannot be combined')
    
    # Chainable properties (override base class to add homogeneity check)
    
    @property
    def ensemble_mean(self):
        """
        Mean over ensemble dimensions (chainable property).
        
        Computes the mean across all series at each time step.
        Requires all series to have compatible sampling rates.
        
        Returns
        -------
        TimeSeries
            Mean across ensemble at each time step (chainable)
        
        Raises
        ------
        Exception
            If series have incompatible sampling rates.
            Use :meth:`is_homogeneous` to check compatibility first.
        
        Examples
        --------
        >>> # Check compatibility before using
        >>> if multi.is_homogeneous():
        ...     mean = multi.ensemble_mean.value
        >>> 
        >>> # Chain with time operations
        >>> overall_mean = multi.ensemble_mean.time_mean.value
        
        See Also
        --------
        is_homogeneous : Check if series have compatible sampling rates
        ensemble_std : Standard deviation across ensemble
        ensemble_median : Median across ensemble
        """
        if not self.is_homogeneous():
            raise Exception('Data series cannot be combined - incompatible sampling rates. '
                          'Use is_homogeneous() to check compatibility.')
        self._impersonateDeltaTime()
        return super().ensemble_mean
    
    @property
    def ensemble_std(self):
        """
        Standard deviation over ensemble dimensions (chainable property).
        
        Computes the std across all series at each time step.
        Requires all series to have compatible sampling rates.
        
        Returns
        -------
        TimeSeries
            Std across ensemble at each time step (chainable)
        
        Raises
        ------
        Exception
            If series have incompatible sampling rates.
            Use :meth:`is_homogeneous` to check compatibility first.
        
        Examples
        --------
        >>> if multi.is_homogeneous():
        ...     std = multi.ensemble_std.value
        >>> 
        >>> # Chain with time operations
        >>> mean_std = multi.ensemble_std.time_mean.value
        
        See Also
        --------
        is_homogeneous : Check if series have compatible sampling rates
        ensemble_mean : Mean across ensemble
        """
        if not self.is_homogeneous():
            raise Exception('Data series cannot be combined - incompatible sampling rates. '
                          'Use is_homogeneous() to check compatibility.')
        self._impersonateDeltaTime()
        return super().ensemble_std
    
    @property
    def ensemble_median(self):
        """
        Median over ensemble dimensions (chainable property).
        
        Computes the median across all series at each time step.
        Requires all series to have compatible sampling rates.
        
        Returns
        -------
        TimeSeries
            Median across ensemble at each time step (chainable)
        
        Raises
        ------
        Exception
            If series have incompatible sampling rates.
            Use :meth:`is_homogeneous` to check compatibility first.
        
        Examples
        --------
        >>> if multi.is_homogeneous():
        ...     median = multi.ensemble_median.value
        >>> 
        >>> # Chain with time operations
        >>> median_avg = multi.ensemble_median.time_mean.value
        
        See Also
        --------
        is_homogeneous : Check if series have compatible sampling rates
        ensemble_mean : Mean across ensemble
        """
        if not self.is_homogeneous():
            raise Exception('Data series cannot be combined - incompatible sampling rates. '
                          'Use is_homogeneous() to check compatibility.')
        self._impersonateDeltaTime()
        return super().ensemble_median
    
    @property
    def ensemble_rms(self):
        """
        RMS over ensemble dimensions (chainable property).
        
        Computes the RMS across all series at each time step.
        Requires all series to have compatible sampling rates.
        
        Returns
        -------
        TimeSeries
            RMS across ensemble at each time step (chainable)
        
        Raises
        ------
        Exception
            If series have incompatible sampling rates.
            Use :meth:`is_homogeneous` to check compatibility first.
        
        Examples
        --------
        >>> if multi.is_homogeneous():
        ...     rms = multi.ensemble_rms.value
        >>> 
        >>> # Chain with time operations
        >>> mean_rms = multi.ensemble_rms.time_mean.value
        
        See Also
        --------
        is_homogeneous : Check if series have compatible sampling rates
        ensemble_mean : Mean across ensemble
        """
        if not self.is_homogeneous():
            raise Exception('Data series cannot be combined - incompatible sampling rates. '
                          'Use is_homogeneous() to check compatibility.')
        self._impersonateDeltaTime()
        return super().ensemble_rms
    
    @property
    def ensemble_ptp(self):
        """
        Peak-to-peak over ensemble dimensions (chainable property).
        
        Computes the peak-to-peak range across all series at each time step.
        Requires all series to have compatible sampling rates.
        
        Returns
        -------
        TimeSeries
            Peak-to-peak across ensemble at each time step (chainable)
        
        Raises
        ------
        Exception
            If series have incompatible sampling rates.
            Use :meth:`is_homogeneous` to check compatibility first.
        
        Examples
        --------
        >>> if multi.is_homogeneous():
        ...     ptp = multi.ensemble_ptp.value
        >>> 
        >>> # Chain with time operations
        >>> mean_ptp = multi.ensemble_ptp.time_mean.value
        
        See Also
        --------
        is_homogeneous : Check if series have compatible sampling rates
        """
        if not self.is_homogeneous():
            raise Exception('Data series cannot be combined - incompatible sampling rates. '
                          'Use is_homogeneous() to check compatibility.')
        self._impersonateDeltaTime()
        return super().ensemble_ptp

# ___oOo___


from itertools import chain  # List flattening

class BaseIndexer():
    """Base indexer for flexible data subset selection.
    
    Indexers provide a flexible interface for selecting subsets of data
    from time series. They process various argument formats:
    
    - Positional arguments for direct element selection
    - Keyword arguments with customizable names
    - Slice syntax for range selection
    
    Parameters
    ----------
    single_kw : str or list of str, optional
        Additional keyword names for single element selection.
        Default keywords: 'element', 'elements'
    from_kw : str or list of str, optional
        Additional keyword names for start of range.
        Default keywords: 'from_element', 'first'
    to_kw : str or list of str, optional
        Additional keyword names for end of range.
        Default keywords: 'to_element', 'last'
    
    Examples
    --------
    >>> indexer = BaseIndexer(single_kw='mode', from_kw='start')
    
    >>> # Select single elements
    >>> idx = indexer.process_args(5)              # Returns 5
    >>> idx = indexer.process_args(element=5)      # Returns 5
    >>> idx = indexer.process_args(mode=5)         # Returns 5
    
    >>> # Select element range
    >>> idx = indexer.process_args(first=10, last=20)  # Returns slice(10, 20)
    >>> idx = indexer.process_args(start=10)           # Returns slice(10, None)
    
    >>> # Select list of elements
    >>> idx = indexer.process_args(elements=[1,5,9])   # Returns [1, 5, 9]
    
    Notes
    -----
    This class is typically used as a base for specialized indexers
    that understand data structure (e.g., x/y slopes, quadrants).
    
    See Also
    --------
    arte.time_series.indexer.Indexer : Advanced indexer with more features
    """

    def __init__(self, single_kw=None, from_kw=None, to_kw=None):
        self.single_kw = ['element', 'elements']
        self.from_kw= ['from_element', 'first']
        self.to_kw = ['to_element', 'last']

        if single_kw:
            self.single_kw += [single_kw]
            self.single_kw = list(chain.from_iterable(self.single_kw))
        if from_kw:
            self.from_kw += [from_kw]
            self.from_kw = list(chain.from_iterable(self.from_kw))
        if to_kw:
            self.to_kw += [to_kw]
            self.to_kw = list(chain.from_iterable(self.to_kw))

    def process_args(self, *args, **kwargs):
        '''
        default: all elements
        element = single element
        elements = list of elements
        from_element = first element
        to_element = last element
        '''
        elements = None
        from_element = None
        to_element = None
        if len(args) > 0:
            elements = args
        else:
            for k, v in kwargs.items():
                if k in self.single_kw:
                    elements = v
                if k in self.from_kw:
                    from_element = v
                if k in self.to_kw:
                    to_element = v
        if elements is not None:
            return elements
        return slice(from_element, to_element)


import numpy as np
from arte.utils.decorator import override
from arte.optical_propagation.abstract_coronograph import Coronograph    

class PerfectCoronograph(Coronograph):
    """ Class to simulate the perfect coronograph 

    Perfect coronograph formula is from equation (1) in:

    Cavarroc, C., Boccaletti, A., Baudoz, P., Fusco, T., and Rouan, D.,
    “Fundamental limitations on Earth-like planet detection with extremely large telescopes”,
    Astronomy and Astrophysics, vol. 447, no. 1, EDP, pp. 397-403, 2006. doi:10.1051/0004-6361:20053916.
    """

    def __init__(self):
        pass

    @override
    def _get_pupil_mask(self, field):
        field_amp = np.abs(field)
        phase = np.angle(field)[field_amp>1e-12]
        phase_var = np.sum((phase-np.mean(phase))**2)/len(phase)
        pupil_mask = field_amp * (np.sqrt(np.exp(-phase_var))\
                                  -np.exp(1j*np.angle(field),dtype=np.complex128)) \
                            * np.exp(-1j*np.angle(field),dtype=np.complex128)
        return pupil_mask
    
    @override
    def _get_focal_plane_mask(self, field):
        return np.ones(field.shape,dtype=bool)



import abc

class AbstractFileNameWalker(metaclass=abc.ABCMeta):
    """Abstract base class for file path management in data analysis.
    
    FileWalker knows the directory structure and file naming conventions
    of your data storage. It provides methods to:
    
    - Locate the root directory of data storage
    - Find the directory for a specific tag
    - Return full paths for each data file type
    - Find tags within a date range
    
    Derived classes must implement snapshot_dir() and typically add
    one method per file type to return complete file paths.
    
    Examples
    --------
    >>> class MyFileWalker(AbstractFileNameWalker):
    ...     def snapshot_dir(self, tag):
    ...         root = Path(os.environ['DATA_ROOT'])
    ...         day = tag.split('_')[0]
    ...         return root / day / tag
    ...     
    ...     def camera_frames(self, tag):
    ...         return self.snapshot_dir(tag) / 'frames.fits'
    ...     
    ...     def dm_commands(self, tag):
    ...         return self.snapshot_dir(tag) / 'dm_commands.npy'
    
    >>> walker = MyFileWalker()
    >>> frames_file = walker.camera_frames('20240101_120000')
    
    See Also
    --------
    BaseAnalyzer : Uses FileWalker to locate data files
    """

    @abc.abstractmethod
    def snapshot_dir(self, tag):
        '''
        Abstract method that must be reimplemented in the derived class.

        Parameters
        ----------
        tag: str
            Unique tag that identifies an analyzer data directory

        Returns
        -------
        fullpath: str
            Full path including the tag directory
        '''
        pass

    def find_tag_between_dates(self, tag_start, tag_stop):
        '''Return the list of tags between the two extremes'''
        return []



import numpy as np
from astropy import units as u
from arte.photometry.filters import Filters
from arte.photometry.normalized_star_spectrum import get_normalized_star_spectrum
from arte.photometry.eso_sky_calc import EsoSkyCalc
from synphot.models import Empirical1D
from synphot.spectrum import SpectralElement
from synphot.observation import Observation
import warnings


def _remove_empty_bins(waveset, f_source, filt):
    """
    Remove bins from waveset that don't contain any samples from source or filter.
    
    This prevents ZeroDivisionError in synphot's calcbinflux when bins are empty.
    
    Parameters
    ----------
    waveset : astropy.units.Quantity
        Input wavelength set
    f_source : synphot.spectrum.SourceSpectrum
        Source spectrum
    filt : synphot.spectrum.SpectralElement
        Filter/bandpass
        
    Returns
    -------
    waveset : astropy.units.Quantity
        Sanitized wavelength set with empty bins removed
    """
    from synphot import binning, utils
    
    ws = u.Quantity(waveset).to(u.Angstrom)
    
    # Calculate bin edges
    bin_edges = binning.calculate_bin_edges(ws.value)
    
    # Merge all wavelengths from bins, source, and filter
    spwave = utils.merge_wavelengths(bin_edges.value, ws.value)
    if f_source.waveset is not None:
        spwave = utils.merge_wavelengths(spwave, f_source.waveset.to(u.Angstrom).value)
    if filt.waveset is not None:
        spwave = utils.merge_wavelengths(spwave, filt.waveset.to(u.Angstrom).value)
    
    # Find which bins contain at least one sample
    idx = np.searchsorted(spwave, bin_edges)
    non_empty = idx[1:] > idx[:-1]
    
    # Keep only non-empty bins
    waveset_clean = ws[non_empty]
    
    if len(waveset_clean) < 2:
        raise ValueError(
            f"Too few non-empty bins in waveset. Original bins: {len(ws)}, "
            f"Non-empty bins: {len(waveset_clean)}")
    
    n_removed = len(ws) - len(waveset_clean)
    if n_removed > 0:
        warnings.warn(
            f"Removed {n_removed} empty bins from waveset "
            f"({n_removed/len(ws)*100:.1f}%)")
    
    return waveset_clean


def expected_flux(
        transmissive_system,
        star_spectral_type='K5V',
        magnitude=0,
        filter_name=Filters.ESO_ETC_R,
        waveset=None,
        return_details=False,
        zenith_angle=30 * u.deg,
        include_sky=True):
    '''
    Estimate the expected flux for the given transmissive system
    from a source of a given spectral type .
    
    Parameters
    ----------
    transmissive_system: arte.photometry.transmissive_elements.TransmissiveSystem or TransmissiveElement
        The optical system through which to calculate the flux
    star_spectral_type: str
        Spectral type of the star (default is 'K5V')
    magnitude: float
        Magnitude of the source in the Vega system (default is 0)
    filter_name: str
        Filter name for magnitude normalization (default is ESO_ETC_R)
    waveset: astropy.units.Quantity, optional
        Wavelength set where to compute the expected flux.
        If None, uses the native waveset of the source spectrum (default is None)
    return_details: bool
        If True, also return f_source and obs_source (default is False)
    zenith_angle: astropy.units.Quantity
        Zenith angle for airmass calculation (default is 30 degrees)
    include_sky: bool
        If True, include sky transmission in the calculation (default is True)
    
    Returns
    -------
    counts_source : astropy.units.Quantity
        Expected flux in e-/s/m^2
    f_source : synphot.spectrum.SourceSpectrum (optional)
        Normalized star spectrum of the source.
        Returned if return_details is True
    obs_source : synphot.observation.Observation (optional)
        Observation object for the source filtered by the transmissive system.
        Returned if return_details is True
    
    Examples
    --------
    Basic usage with default parameters:
    
    >>> from arte.photometry.transmissive_elements_catalogs import FiltersCatalog
    >>> filt = FiltersCatalog.bessel_H()
    >>> flux = expected_flux(filt)
    
    K5V star with magnitude 8 in R filter through Bessel H:
    
    >>> flux = expected_flux(filt, star_spectral_type='K5V', magnitude=8, 
    ...                      filter_name=Filters.ESO_ETC_R)
    
    Custom wavelength set from 1500 to 1700 nm and zenith angle:
    
    >>> wv = np.arange(15000, 17000, 50) * u.Angstrom
    >>> flux = expected_flux(filt, waveset=wv, zenith_angle=45*u.deg)
    
    Get detailed output and exclude sky contribution:
    
    >>> flux, f_src, obs = expected_flux(filt, return_details=True, include_sky=False)
    
    Compare flux in K band for 2 different glasses:
    
    >>> wvset = np.arange(2000, 2200, 10) * u.nm
    >>> ideal = expected_flux(TransmissiveElement.ideal(), waveset=wvset)
    >>> suprasil = expected_flux(GlassesCatalog.suprasil3002_10mm_internal_001(), waveset=wvset)
    >>> npsk53a = expected_flux(GlassesCatalog.schott_NPSK53A_10_mm_internal_001(), waveset=wvset)
    '''
    
    f_source = get_normalized_star_spectrum(spectral_type=star_spectral_type,
                                            magnitude=magnitude,
                                            filter_name=filter_name)

    airmass = 1 / np.cos(zenith_angle.to(u.rad))
    
    if include_sky:
        sky = EsoSkyCalc(airmass=airmass, incl_moon='N')
        sky_se = SpectralElement(Empirical1D, points=sky.lam,
                                 lookup_table=sky.trans)
        filt = sky_se * transmissive_system.transmittance
    else:
        filt = transmissive_system.transmittance

    warnings.filterwarnings('ignore')

    # Generate initial binset if not provided
    if waveset is None:
        # Let synphot determine the binset
        if filt.waveset is not None:
            initial_binset = filt.waveset
        elif f_source.waveset is not None:
            initial_binset = f_source.waveset
        else:
            raise ValueError("Both source and filter have undefined waveset")
        
        # Remove empty bins from the generated binset
        waveset = _remove_empty_bins(initial_binset, f_source, filt)
    else:
        # Remove empty bins from user-provided waveset
        waveset = _remove_empty_bins(waveset, f_source, filt)

    obs_source = Observation(spec=f_source, band=filt,
                             binset=waveset, force='taper')

    area_subap = 1 * u.m ** 2
    exp_time = 1 * u.s
    
    counts_source = obs_source.countrate(area=area_subap, binned=False, wavelengths=waveset)
    counts_source = (counts_source * exp_time).decompose()  # e-/s/m^2

    print('Expected flux for %s star (mag=%s in %s filter) in e-/s/m2 (x1e8): %s'
          % (star_spectral_type, magnitude, filter_name, counts_source / 1e8))

    if return_details:
        return counts_source, f_source, obs_source
    return counts_source



from warnings import warn
from arte.utils.zernike_decomposer import ZernikeModalDecomposer


class ModalDecomposer(ZernikeModalDecomposer):
    '''
    Backward compatibility with old ModalDecomposer class
    '''
    def __init__(self, n_modes=None, n_zernike_modes=None):
        if n_zernike_modes is not None:
            n_modes = n_zernike_modes
        if n_modes is None and n_zernike_modes is None:
            raise ValueError("either n_modes or n_zernike_modes must be specified")
        super().__init__(n_modes)
        warn(
            f"{self.__class__.__name__} will be deprecated or change use. Use ZernikeModalDecomposer instead",
            DeprecationWarning,
            stacklevel=2,
        )




import datetime
import functools

from arte.utils.help import add_help
from arte.dataelab.cache_on_disk import set_tag, clear_cache


class PostInitCaller(type):
    '''Meta class with a _post_init() method.
    
    Used to initialize the disk cache, since DiskCacher objects are
    available only after all child/member objects have been initialized'''
    def __call__(cls, *args, **kwargs):
        obj = type.__call__(cls, *args, **kwargs)
        obj._post_init()
        return obj


@add_help
class BaseAnalyzer(metaclass=PostInitCaller):
    """Main analyzer object for adaptive optics data analysis.
    
    The Analyzer is the primary interface for loading and analyzing data
    identified by a unique tag (typically a timestamp). It coordinates:
    
    - Data file discovery via FileWalker
    - Lazy loading of time series data
    - Caching of computed results to disk
    - Access to raw and processed data streams
    
    Analyzers are typically accessed via the get() classmethod which maintains
    an internal cache, ensuring that multiple requests for the same tag return
    the same instance.
    
    Parameters
    ----------
    snapshot_tag : str
        Unique identifier for this dataset (e.g., '20240101_120000')
    recalc : bool, optional
        If True, clear all cached data and recompute on access (default: False)
    
    Attributes
    ----------
    _snapshot_tag : str
        The tag identifying this dataset
    
    Examples
    --------
    >>> analyzer = MyAnalyzer.get('20240101_120000')
    >>> modes = analyzer.residual_modes.get_data()
    >>> std_per_mode = analyzer.residual_modes.time_std()
    
    >>> # Force recalculation of cached data
    >>> analyzer = MyAnalyzer.get('20240101_120000', recalc=True)
    
    Notes
    -----
    When creating a derived class:
    
    1. Define data streams as attributes (e.g., self.residual_modes)
    2. Use DataLoaders for lazy file loading
    3. Methods returning expensive computations can use @cache_on_disk
    4. The _post_init() method is called automatically after __init__
    
    See Also
    --------
    BaseAnalyzerSet : For analyzing multiple tags together
    cache_on_disk : Decorator for persistent caching
    """

    def __init__(self, snapshot_tag, recalc=False):
        self._snapshot_tag = snapshot_tag
        self._recalc = recalc
 
    def _post_init(self):
        '''Initialize disk cache
        
        Executed after all child classes have completed their __init__
        thanks to the PostInitCaller metaclass
        '''
        set_tag(self, self._snapshot_tag)
        if self._recalc:
            self.recalc()
            self._recalc = False

    @classmethod
    def get(cls, tag, *args, recalc=False, **kwargs):
        '''Get the Analyzer instance (or derived class) corresponding to *tag*.

        This method mantains an internal cache. If a tag is requested
        multiple times, the same Analyzer instance is returned.
        
        Parameters
        ----------
        tag: str
            snapshot tag
        recalc: bool, optional
            if set to True, any cached data for this tag will be deleted and
            computed again when requested
         '''
        analyzer = cls._get(tag, *args, **kwargs)
        # Special recalc handling, must be set again for a cached instance
        if recalc:
            analyzer.recalc()
        return analyzer

    @classmethod
    @functools.cache
    def _get(cls, tag, *args, recalc=False, **kwargs):
        '''Get a new Analyzer instance 

        Added one level of indirection (get() calls _get())
        to make sure that the cache always works even when
        the default argument *recalc* is not specified in get().

        Also makes it easier to override it in classes
        '''
        return cls(tag, *args, recalc=recalc, **kwargs)

    def recalc(self):
        '''Force recalculation of this analyzer data'''
        clear_cache(self)

    def snapshot_tag(self):
        '''Snapshot tag for this Analyzer object'''
        return self._snapshot_tag

    def date_in_seconds(self):
        '''Tag date as seconds since the epoch'''
        epoch = datetime.datetime(1970, 1, 1, 0, 0, 0, 0)
        this_date = datetime.datetime(int(self._snapshot_tag[0:4]),
                                     int(self._snapshot_tag[4:6]),
                                     int(self._snapshot_tag[6:8]),
                                     int(self._snapshot_tag[9:11]),
                                     int(self._snapshot_tag[11:13]),
                                     int(self._snapshot_tag[13:15]))
        td = this_date - epoch
        return (td.microseconds + (td.seconds + td.days * 24 * 3600) * 1e6) / 1e6

    # Override to add additional info
    def _info(self):
        return {'snapshot_tag': self._snapshot_tag}

    def info(self):
        '''Info dictionary'''
        return self._info()

    def summary(self, keywidth=None):
        '''Print info dictionary on stdout'''
        info = self._info()
        if keywidth is None:
            keywidth = max(len(x) for x in info.keys())
        for k, v in info.items():
            print(f'{k:{keywidth}} : {v}')

    def wiki(self, header=True):
        '''Print info dictionary in wiki format on stdout'''
        info = self._info()
        spacing = [max([len(x)+2, len(info[x])]) for x in info.keys()]

        if header:
            for i, k in enumerate(info.keys()):
                print(f'|*{k:^{spacing[i]}}*', end='')
            print('|')

        for i, v in enumerate(info.values()):
            print(f'|{v:^{spacing[i]}}', end='')
        print('|')




import numpy as np
from arte.atmo.abstract_phase_screen_generator import AbstractPhaseScreenGenerator
from astropy.io import fits
from arte.utils.decorator import override


class PhaseScreenGenerator(AbstractPhaseScreenGenerator):
    """ 
    Class for atmospheric phase screens generation

    Example use:
    >>> phs = PhaseScreenGenerator(screenSizeInPixels=256,
    ...                            screenSizeInMeters=10.0,
    ...                            outerScaleInMeters=25.0,
    ...                            seed=42)
    >>> phs.generate_normalized_phase_screens(numberOfScreens=5)
    >>> phs.rescale_to(15e-2)  # r0 at 500nm = 15 cm
    >>> phaseScreensInM = phs.get_in_meters()
    >>> phaseScreensAt1um = phs.get_in_radians_at(1e-6)
    """


    def __init__(self,
                 screenSizeInPixels,
                 screenSizeInMeters,
                 outerScaleInMeters,
                 seed:int=None,
                 nSubHarmonics:int=8):
        super().__init__(screenSizeInPixels,screenSizeInMeters,seed,nSubHarmonics)
        self._outerScaleInM = float(outerScaleInMeters)

    @override
    def _get_power_spectral_density(self, freqMap):
        """ Von Karman PSD """
        mappa = (freqMap**2 + (self._screenSzInM / self._outerScaleInM)**2)**(
            -11. / 12)
        mappa[0, 0] = 0
        return mappa
    
    @override
    def _get_scaling(self):
        """ Scaling factor for Von Karman spectrum """
        return np.sqrt(0.0228) * self._screenSzInPx**(5. / 6)


    def rescale_to(self, r0At500nm):
        self._normalizationFactor = \
            ((self._screenSzInM / self._screenSzInPx) / r0At500nm) ** (5. / 6)

    def get_in_radians_at(self, wavelengthInMeters):
        return 500e-9 / wavelengthInMeters * \
            self._normalizationFactor * self._phaseScreens

    def get_in_meters(self):
        return self._normalizationFactor * self._phaseScreens / \
            (2 * np.pi) * 500e-9
    
    def save_normalized_phase_screens(self, filepath:str, overwrite:bool=False):
        hdr = fits.Header()
        hdr['SZ_IN_PX'] =  self._screenSzInPx
        hdr['SZ_IN_M'] = self._screenSzInM
        hdr['OS_IN_M'] = self._outerScaleInM
        hdr['SEED'] = self._seed
        hdr['NSUBHARM'] = self._nSubHarmonicsToUse
        fits.writeto(filepath, self._phaseScreens, hdr, overwrite=overwrite)
        
    @staticmethod 
    def load_normalized_phase_screens(filepath:str):
        header = fits.getheader(filepath)
        screenSizeInPixels = header['SZ_IN_PX'] 
        screenSizeInMeters = header['SZ_IN_M'] 
        outerScaleInMeters = header['OS_IN_M'] 
        seed = header['SEED'] 
        try: # added for retro-compatibility
            nSubHarmonics = header['NSUBHARM']
        except KeyError:
            nSubHarmonics = 8
        psg = PhaseScreenGenerator(
            screenSizeInPixels,
            screenSizeInMeters,
            outerScaleInMeters,
            nSubHarmonics,
            seed)
        hduList = fits.open(filepath)
        psg._phaseScreens = hduList[0].data
        return psg
